import argparse
import matplotlib.pyplot as plt
import numpy as np
import sys

from . import histo


def argument_parser():
    p = argparse.ArgumentParser(description="Clustering pixel value histogram")
    p.add_argument("-n", default=2, type=int, help="Number of component")
    p.add_argument("-M", "--method", default="gmm", help="(kmeanss|gmm)")
    p.add_argument("-N", "--noise", metavar="SD", type=float, default=None, help="Additive gaussian noise")
    p.add_argument("-p", "--pixelvalue-range", metavar="RANGE", action=histo.RangeAction)
    p.add_argument("input", help="Input image file name")
    p.add_argument("output", nargs="?", default=None, help="Histogram output file")
    return p


def main(args):
    from .histo import load_pixels, gmm_plot, kmeans_plot, dpgmm_plot, vbgmm_plot, dbscan_plot

    pixels = load_pixels(args.input)

    if args.noise is not None:
        pixels = pixels + args.noise * np.random.randn(*pixels.shape)

    if args.pixelvalue_range:
        pixels = histo.filter_by_range(pixels, *args.pixelvalue_range)

    pixels = pixels.reshape((pixels.size, 1))

    plt.xlim(0, 280)
    plt.xlabel("Pixel value")
    plt.ylabel("Number of pixels")
    plt.hist(pixels, bins=280, range=(0, 279))

    if args.method == "gmm":
        gmm_plot(pixels, args.n)
    elif args.method == "kmeans":
        kmeans_plot(pixels, args.n)
    elif args.method == "dpgmm":
        dpgmm_plot(pixels, args.n)
    elif args.method == "vbgmm":
        vbgmm_plot(pixels, args.n)
    elif args.method == "DBSCAN":
        dbscan_plot(pixels)
    elif args.method == "nofit":
        pass
    else:
        print("Unknown method: %s" % args.method)
        sys.exit(1)

    if args.output is not None:
        plt.savefig(args.output)
    else:
        plt.show()

    return 0


if __name__ == "__main__":
    sys.exit(main(argument_parser().parse_args()))



import sys
from argparse import ArgumentParser

import numpy as np
import matplotlib.pyplot as plt

from homcloud.argparse_common import add_arguments_for_load_diagrams
from homcloud.version import __version__
import homcloud.pdgm as pdgm


def main(args=None):
    args = args or argument_parser().parse_args()
    pd = pdgm.load_merged_diagrams(args.input, args.type, args.degree)
    transl, mat = transform_to_x_axis(np.array([args.birth1, args.death1]), np.array([args.birth2, args.death2]))
    xy = np.dot(mat, np.array([pd.births, pd.deaths]) - transl.reshape(2, 1))
    mask = (xy[0, :] >= 0) & (xy[0, :] <= 1) & (np.abs(xy[1, :]) < args.width / 2)
    histo_input = xy[0, mask]
    (hist, *_) = plt.hist(histo_input, bins=args.bins, range=(0, 1), log=args.log)
    plt.xlim(0, 1)
    plt.xticks([0, 1])
    ax = plt.gca()
    ax.set_xticklabels(
        [
            construct_label(args.left_label, args.birth1, args.death1),
            construct_label(args.right_label, args.birth2, args.death2),
        ]
    )
    if args.title is not None:
        ax.set_title(args.title)
    if args.output:
        if args.text_output:
            np.savetxt(args.output, hist, fmt="%d")
        else:
            plt.savefig(args.output, dpi=args.dpi)
    else:
        plt.show()


def argument_parser():
    p = ArgumentParser(description="")
    p.add_argument("-V", "--version", action="version", version=__version__)
    add_arguments_for_load_diagrams(p)
    p.add_argument("-l", "--left-label", help="label for starting point")
    p.add_argument("-r", "--right-label", help="label for end point")
    p.add_argument("-b", "--bins", type=int, default=100, help="number of bins")
    p.add_argument("--log", default=False, action="store_true", help="log scale")
    p.add_argument("-o", "--output", help="output file path")
    p.add_argument("--text-output", action="store_true", default=False, help="output histgram data into a text file")
    p.add_argument("-t", "--title", help="title")
    p.add_argument("--dpi", type=int, help="output dpi")
    p.add_argument("birth1", type=float, help="birth of starting point")
    p.add_argument("death1", type=float, help="death of starting point")
    p.add_argument("birth2", type=float, help="birth of end point")
    p.add_argument("death2", type=float, help="death of end point")
    p.add_argument("width", type=float, help="width for histogram")
    p.add_argument("input", nargs="+", help="input file path")
    return p


def transform_to_x_axis(v1, v2):
    d = v2 - v1
    norm = np.linalg.norm(d)
    rotation = np.array([[d[0], d[1]], [-d[1], d[0]]]) / norm
    scale_x = np.array(([1 / norm, 0], [0, 1]))
    return v1, np.dot(scale_x, rotation)


def construct_label(labelstr, birth, death):
    if labelstr is None:
        return "({:4f},{:4f})".format(birth, death)
    else:
        return labelstr


if __name__ == "__main__":
    sys.exit(main())



import argparse
import sys

from homcloud.version import __version__

from homcloud.optvol import sign_function
from homcloud.license import add_argument_for_license
import homcloud.int_reduction_ext as int_reduction_ext
from homcloud.pdgm import PDGM


def build_checker(dim, boundary_map):
    checker = int_reduction_ext.Checker(count_num_cells(dim, boundary_map))
    setup_matrix(checker, boundary_map)
    return checker


def setup_matrix(matrix, boundary_map):
    sign = sign_function(boundary_map)
    map = boundary_map["map"]

    def dim(i):
        return map[i][0]

    def column(i):
        return map[i][1]

    for i in range(len(map)):
        matrix.add_cell(dim(i))
        signed_column = [(j, sign(i, kth)) for kth, j in enumerate(column(i))]
        for j, s in sorted(signed_column):
            matrix.add_boundary_coef(i, j, s)


def count_num_cells(dim, boundary_map):
    num_cells = [0] * (dim + 1)
    for d, *_ in boundary_map["map"]:
        num_cells[d] += 1
    return num_cells


def main(args=None):
    args = args or argument_parser().parse_args()
    diagram = PDGM.open(args.input, 0, False)
    checker = build_checker(diagram.input_dim, diagram.boundary_map_chunk)

    stop_coef, stop_index = checker.check()
    if stop_coef == 0:
        print("No Problem")
        return 0
    else:
        stop_time = diagram.index_to_level[stop_index]
        print("Problematic coef: {}, Problematic time: {}".format(stop_coef, stop_time))
        return 1


def argument_parser():
    p = argparse.ArgumentParser(description="Examine field problem")
    p.add_argument("-V", "--version", action="version", version=__version__)
    add_argument_for_license(p)
    p.add_argument("input", help="input file")
    return p


if __name__ == "__main__":
    sys.exit(main())



try:
    from tempfile import TemporaryDirectory as TD
except ImportError:
    from backports.tempfile import TemporaryDirectory as TD

try:
    from unittest.mock import Mock as Mk
except ImportError:
    from mock import Mock as Mk

try:
    import imageio.v2 as imageio
except ImportError:
    import imageio


try:
    from sklearn.metrics import DistanceMetric
except ImportError:
    from sklearn.neighbors import DistanceMetric


TemporaryDirectory = TD
Mock = Mk

INFINITY = float("inf")



import msgpack
import math
import struct
from pathlib import Path

from cached_property import cached_property
import numpy as np

MAGIC_HEADER = (b"HomCloud .pdgm format" + b"\0" * 24)[0:24]
MAGIC_HEADER_LENGTH = len(MAGIC_HEADER)
HEADER_LENGTH = MAGIC_HEADER_LENGTH + 8
FORMAT_VERSION = 1


class PDChunk(object):
    def __init__(self, chunktype, degree, births, deaths, ess_births):
        self.chunktype = chunktype
        self.degree = degree
        self.births = births
        self.deaths = deaths
        self.ess_births = ess_births

    @cached_property
    def binary(self):
        return msgpack.packb(
            {
                "chunktype": self.chunktype,
                "degree": self.degree,
                "births": self.births,
                "deaths": self.deaths,
                "essential_births": self.ess_births,
            },
            use_bin_type=True,
        )

    def metadata(self, start):
        return {"chunktype": self.chunktype, "degree": self.degree, "start": start, "length": len(self.binary)}


class SimpleChunk(object):
    def __init__(self, chunktype, data):
        self.chunktype = chunktype
        self.data = data

    @cached_property
    def binary(self):
        return msgpack.packb({"chunktype": self.chunktype, "data": self.data}, use_bin_type=True)

    def metadata(self, start):
        return {"chunktype": self.chunktype, "start": start, "length": len(self.binary)}


class BinaryChunk(object):
    def __init__(self, chunktype, binary):
        self.chunktype = chunktype
        self.binary = binary

    def metadata(self, start):
        return {"chunktype": self.chunktype, "start": start, "length": len(self.binary)}


class AllPairsChunk(object):
    def __init__(self, degree, pairs):
        self.degree = degree
        self.pairs = pairs

    @cached_property
    def binary(self):
        return msgpack.packb({"chunktype": "allpairs", "degree": self.degree, "data": self.pairs})

    def metadata(self, start):
        return {"chunktype": "allpairs", "degree": self.degree, "start": start, "length": len(self.binary)}


class SimpleDegreeChunk(object):
    def __init__(self, chunktype, degree, data):
        self.chunktype = chunktype
        self.degree = degree
        self.data = data

    @cached_property
    def binary(self):
        return msgpack.packb({"chunktype": self.chunktype, "degree": self.degree, "data": self.data})

    def metadata(self, start):
        return {"chunktype": self.chunktype, "degree": self.degree, "start": start, "length": len(self.binary)}


class BitmapInformationChunk(object):
    def __init__(self, shape, periodicity):
        self.shape = shape
        self.periodicity = periodicity

    @cached_property
    def binary(self):
        return msgpack.packb(
            {
                "chunktype": "bitmap_information",
                "shape": self.shape,
                "periodicity": self.periodicity,
            }
        )

    def metadata(self, start):
        return {"chunktype": "bitmap_information", "start": start, "length": len(self.binary)}


class AlphaInformationChunk(object):
    def __init__(self, num_vertices, periodicity, weighted, squared):
        self.num_vertices = num_vertices
        self.periodicity = periodicity
        self.weighted = weighted
        self.squared = squared

    @cached_property
    def binary(self):
        return msgpack.packb(
            {
                "chunktype": "alpha_information",
                "num_vertices": self.num_vertices,
                "periodicity": self.periodicity,
                "weighted": self.weighted,
                "squared": self.squared,
            }
        )

    def metadata(self, start):
        return {"chunktype": "alpha_information", "start": start, "length": len(self.binary)}


class RipsInformationChunk(object):
    def __init__(self, num_vertices):
        self.num_vertices = num_vertices

    @cached_property
    def binary(self):
        return msgpack.packb({"chunktype": "rips_information", "num_vertices": self.num_vertices})

    def metadata(self, start):
        return {"chunktype": "rips_information", "start": start, "length": len(self.binary)}


class BoundaryMapChunk(object):
    def __init__(self, type, map):
        self.type = type
        self.map = map

    @cached_property
    def binary(self):
        return msgpack.packb({"chunktype": "boundary_map", "type": self.type, "map": self.map})

    def metadata(self, start):
        return {"chunktype": "boundary_map", "start": start, "length": len(self.binary)}


class GraphWeightsChunk(object):
    def __init__(self, matrix):
        assert matrix.ndim == 2 and matrix.shape[0] == matrix.shape[1]
        self.matrix = matrix

    @cached_property
    def binary(self):
        return msgpack.packb(
            {
                "chunktype": "graph_weights",
                "num_vertices": self.num_vertices,
                "weights": self.graph_weights(),
            }
        )

    def graph_weights(self):
        return self.matrix[np.triu_indices(self.num_vertices, 1)].tolist()

    @property
    def num_vertices(self):
        return self.matrix.shape[0]

    def metadata(self, start):
        return {"chunktype": "graph_weights", "start": start, "length": len(self.binary)}


class PDGMWriter(object):
    def __init__(self, out, filtration_type, dim=None):
        self.out = out
        self.filtration_type = filtration_type
        self.sign_flipped = False
        self.dim = dim
        self.chunks = []

    def append_chunk(self, chunk):
        self.chunks.append(chunk)

    def extend_chunks(self, chunks):
        self.chunks.extend(chunks)

    def append_simple_chunk(self, chunktype, data, degree=None):
        if degree is None:
            self.chunks.append(SimpleChunk(chunktype, data))
        else:
            self.chunks.append(SimpleDegreeChunk(chunktype, degree, data))

    def write(self):
        self.write_magic_header()
        self.write_metadata()
        self.write_chunks()

    def write_magic_header(self):
        self.out.write(MAGIC_HEADER)

    def write_metadata(self):
        metadata_binary = msgpack.packb(self.metadata(), use_bin_type=True)
        self.out.write(struct.pack("q", len(metadata_binary)))
        self.out.write(metadata_binary)

    def write_chunks(self):
        for chunk in self.chunks:
            self.out.write(chunk.binary)

    def metadata(self):
        return {
            "format_version": FORMAT_VERSION,
            "filtration_type": self.filtration_type,
            "sign_flipped": self.sign_flipped,
            "dim": self.dim,
            "chunks": self.chunk_metadata_list(),
        }

    def chunk_metadata_list(self):
        start = 0
        metadata_list = []
        for chunk in self.chunks:
            metadata = chunk.metadata(start)
            start += metadata["length"]
            metadata_list.append(metadata)

        return metadata_list

    def save_pairs(self, pairs, index_to_levels, save_index_info=True):
        indexed_pds = [IndexedPD(d, index_to_levels) for d in range(self.dim + 1)]
        for d, birth, death in pairs:
            if d > self.dim:
                continue
            indexed_pds[d].append(birth, death)

        self.chunks.extend(indexed_pd.to_unindexed_chunk() for indexed_pd in indexed_pds)

        if not save_index_info:
            return

        self.chunks.extend(indexed_pd.to_chunk() for indexed_pd in indexed_pds)
        self.chunks.extend(indexed_pd.to_allpairs_chunk() for indexed_pd in indexed_pds)


class IndexedPD(object):
    def __init__(self, degree, index_to_levels):
        self.degree = degree
        self.birth_indices = []
        self.death_indices = []
        self.ess_birth_indices = []
        self.births = []
        self.deaths = []
        self.ess_births = []
        self.pairs = []
        self.index_to_levels = index_to_levels

    def append(self, birth, death):
        self.pairs.append((birth, death))
        birth_level = self.level(birth)
        death_level = self.level(death)

        if birth_level == -math.inf:
            pass
        elif birth_level == death_level:
            pass
        elif death is None or math.isinf(death_level):
            self.ess_birth_indices.append(birth)
            self.ess_births.append(birth_level)
        else:
            self.birth_indices.append(birth)
            self.births.append(birth_level)
            self.death_indices.append(death)
            self.deaths.append(death_level)

    def level(self, index):
        return self.index_to_levels[index] if index is not None else None

    def to_chunk(self):
        return PDChunk("indexed_pd", self.degree, self.birth_indices, self.death_indices, self.ess_birth_indices)

    def to_unindexed_chunk(self):
        return PDChunk("pd", self.degree, self.births, self.deaths, self.ess_births)

    def to_allpairs_chunk(self):
        return AllPairsChunk(self.degree, self.pairs)


class PDGMReader(object):
    def __init__(self, infile, path=None):
        self.infile = infile
        self.path = path
        self.scan_magic_header()
        self.load_metadata()

    @staticmethod
    def open(path):
        return PDGMReader(open(path, "rb"), path)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def scan_magic_header(self):
        if self.infile.read(MAGIC_HEADER_LENGTH) != MAGIC_HEADER:
            raise RuntimeError("HomCloud .pdgm format magic header not found")

    def load_metadata(self):
        (metadata_length,) = struct.unpack("q", self.infile.read(8))
        self.metadata = msgpack.unpackb(self.infile.read(metadata_length), raw=False)
        self.pos_first_chunk = HEADER_LENGTH + metadata_length

    @property
    def filtration_type(self):
        return self.metadata["filtration_type"]

    def unpack_chunk(self, chunk_metadata):
        return msgpack.unpackb(self.read_chunk_bytes(chunk_metadata), raw=False)

    def read_chunk_bytes(self, chunk_metadata):
        self.infile.seek(self.pos_first_chunk + chunk_metadata["start"])
        return self.infile.read(chunk_metadata["length"])

    def load_pd_chunk(self, chunktype, degree):
        chunk = self.load_chunk(chunktype, degree)
        if chunk:
            return chunk["births"], chunk["deaths"], chunk["essential_births"]
        else:
            return (None, None, None)

    def load_simple_chunk(self, chunktype, degree=None):
        chunk = self.load_chunk(chunktype, degree)
        if chunk:
            return chunk["data"]
        else:
            return None

    def load_boundary_map_chunk(self):
        return self.load_chunk("boundary_map")

    def load_chunk(self, chunktype, degree=None):
        chunk_metadata = self.find_chunk_metadata(chunktype, degree)
        if not chunk_metadata:
            return None
        return self.unpack_chunk(chunk_metadata)

    def load_chunk_bytes(self, chunktype, degree=None):
        chunk_metadata = self.find_chunk_metadata(chunktype, degree)
        if not chunk_metadata:
            return None
        return self.read_chunk_bytes(chunk_metadata)

    def find_chunk_metadata(self, chunktype, degree=None):
        def cond(chunk):
            return (chunk["chunktype"] == chunktype) and ((degree is None) or (chunk["degree"] == degree))

        return next(filter(cond, self.metadata["chunks"]), None)

    def load_bitmap_information_chunk(self):
        chunk = self.load_chunk("bitmap_information")
        if chunk:
            return BitmapInformationChunk(chunk["shape"], chunk["periodicity"])
        else:
            return None

    def close(self):
        self.infile.close()


def ispdgm(path):
    return Path(path).suffix == ".pdgm"



import homcloud.pict.distance_transform as pict_distance_transform


def distance_transform(pict, signed=False, metric="manhattan", periodicity=None, obstacle=None, VI=None, V=None):
    """Distance transform.

    Args:
        pict (numpy.ndarray): A binary picture. dtype must be bool.
        signed (bool): True for signed distance transform
        metric (string): Metric. The following metrics are available:
            "manhattan", "chebyshev", "euclidean", "mahalanobis"
        periodicity (None or list of bool): Periodic boundary condition
        obstacle (None or numpy.ndarray): Obstacle bitmap
        VI (None or numpy.ndarray): Inverse matrix for Mahalanobis metric
        V (None or numpy.ndarray): Matrix for Mahalanobis metric

    Returns:
        An ndarray object.

    Example:
        >>> import homcloud.interface as hc
        >>> import numpy as np
        >>> bitmap = np.array([[0, 1, 0, 0], [1, 1, 0, 1]], dtype=bool)
        >>> hc.PDList.from_bitmap_levelset(hc.distance_transform(bitmap, True))
        -> Returns a new PDList

    Remarks:
        The current implementation for periodic BC is a simple periodic image
        copy method. Hence the performance is not so good.
        The developers of HomCloud plan to implove the efficiency
        in the future.
    """
    assert pict.dtype == bool
    return pict_distance_transform.distance_transform(
        pict, metric, signed, periodicity=periodicity, obstacle=obstacle, VI=VI, V=V
    )



from homcloud.visualize_3d import ParaViewSparseCubeDrawer, ParaViewSimplexDrawer
import homcloud.paraview_interface as pv_interface
import homcloud.plotly_3d as p3d


class Optimal1Cycle(object):
    """The class represents an optimal (not volume-optimal) 1-cycle.

    This class is available only for alpha, cubical, rips, and abstract filtration
    with boundary map information.

    You can aquaire an optimal one cycle by :meth:`Pair.optimal_1_cycle`.
    """

    def __init__(self, pair, path):
        self.pair = pair
        self.path_edges = path

    def birth_time(self):
        """Returns the birth time of the pair.

        Returns:
            float: The birth time
        """
        return self.pair.birth_time()

    def death_time(self):
        """Returns the death time of the pair.

        Returns:
            float: The death time
        """
        return self.pair.death_time()

    def birth_position(self):
        """Birth position of the birth-death pair"""
        return self.pair.birth_position

    def path(self):
        """
        Returns:
            list of cell: All edges in the optimal 1-cycle.
        """
        return self.geometry_resolver(False).resolve_cells(self.path_edges)

    def path_symbols(self):
        """
        Returns:
            list of ssimplex: All edges in the optimal 1-cycle
            in the form of the symbolic representation.
        """
        return self.geometry_resolver(True).resolve_cells(self.path_edges)

    def boundary_points(self):
        """
        Returns:
            list of point: All vertices in the optimal 1-cycle.
        """
        return self.geometry_resolver(False).resolve_vertices(self.path_edges)

    def boundary_points_symbols(self):
        """
        Returns:
            list of string: All vertices in the optimal 1-cycle
            in the form of the symbolic representation.
        """
        return self.geometry_resolver(True).resolve_vertices(self.path_edges)

    def geometry_resolver(self, use_symbol):
        return self.pair.diagram.pd.geometry_resolver(use_symbol)

    def to_paraview_node(self, gui_name=None):
        geom_resolver = self.geometry_resolver(False)

        if self.pair.diagram.filtration_type == "cubical":
            drawer = ParaViewSparseCubeDrawer(1, 3, {"isboundary": "1"})
        elif self.pair.diagram.filtration_type == "alpha":
            drawer = ParaViewSimplexDrawer(1, geom_resolver.vertices, {"isboundary": "1"})
        else:
            raise RuntimeError("Unsupported filtration type")

        geom_resolver.draw_cells(drawer, self.path_edges, drawer.various_colors[0])

        f = pv_interface.TempFile(".vtk")
        drawer.write(f)
        f.close()
        return pv_interface.VTK(f.name, gui_name, f).set_representation("Wireframe")

    to_pvnode = to_paraview_node

    def to_plotly3d_trace(self, color=None, width=1, name=""):
        """
        Constructs a plotly's trace object to visualize the optimal 1-cycle

        Args:
            color (string or None): The name of the color
            width (int): The width of the lines
            name (string): The name of the object

        Returns:
            plotly.graph_objects.Scatter3d: Plotly's trace object
        """
        if self.filtration_type == "alpha":
            return p3d.Simplices(self.path(), color, width, name)
        elif self.filtration_type == "cubical":
            return p3d.Cubes(self.path(), color, width, name)
        else:
            raise RuntimeError(f"Unsupported filtration type {self.filtration_type} for plotly3d")

    to_plotly3d = to_plotly3d_trace

    def to_pyvista_mesh(self):
        """
        Constructs a PyVista's mesh object to visualize the optimal 1-cycle.

        Returns:
            pyvista.PolyData: PyVista's mesh object
        """
        import homcloud.pyvistahelper as pvhelper

        if self.filtration_type == "alpha":
            return pvhelper.Lines(self.path())
        raise RuntimeError(f"Unsupported filtration type {self.filtration_type} for pyvista")

    @property
    def filtration_type(self):
        return self.pair.diagram.filtration_type



import argparse


LICENSE_TERMS = """
Copyright 2017-2019 Ippei Obayashi
Copyright 2017-2018 Hiraoka lab (AIMR, Tohoku Univ.)
Copyright 2018-2019 Topological Data Analysis Team (AIP, RIKEN)
Copyright 2019 National Institute of Advanced Industrial Science and Technology (AIST)


HomCloud is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HomCloud is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with HomCloud.  If not, see <http://www.gnu.org/licenses/>.
"""


def add_argument_for_license(parser):
    parser.add_argument("--license", action=ShowLicense, nargs=0, help="show license and exit")


class ShowLicense(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        print(LICENSE_TERMS)
        exit(0)



import subprocess
import os
import math
import struct
from tempfile import TemporaryDirectory


def pairs_from_filtration(filtration, parallels=1, dual=False, upper_dim=None, upper_value=None):
    with TemporaryDirectory() as tmpdir:
        complex_path = os.path.join(tmpdir, "tmp.complex")
        diagram_path = os.path.join(tmpdir, "tmp.diagram")
        with open(complex_path, "wb") as f:
            filtration.write_dipha_complex(f)
        run(complex_path, diagram_path, parallels, dual, upper_dim, upper_value)

        with open(diagram_path, "rb") as f:
            yield from read_pairs(f)


def read_pairs(f):
    f.read(16)  # skip header
    (num_pairs,) = struct.unpack("q", f.read(8))

    for _ in range(num_pairs):
        d, birth, death = struct.unpack("qdd", f.read(24))
        if d < 0:
            yield (-d - 1, birth, math.inf)
        else:
            yield (d, birth, death)


def run(inpath, outpath, parallels=1, dual=False, upper_dim=None, upper_value=None):
    options = []

    if os.environ.get("HOMCLOUD_SUPPRESS_MPI") == "1":
        mpi = []
    else:
        mpi = ["mpiexec", "-n", str(parallels)]
    if dual:
        options.append("--dual")
    if upper_dim is not None:
        options.extend(["--upper_dim", str(upper_dim + 1)])
    if upper_value is not None and upper_value != math.inf:
        options.extend(["--upper_value", str(upper_value)])

    subprocess.check_call(mpi + ["dipha"] + options + [inpath, outpath])



import sys
import os
import argparse
import tempfile
import uuid
import shlex
import functools
import json

from PyQt5.QtWidgets import QMainWindow, QApplication, QMessageBox, QDialog, QProgressDialog
from PyQt5.QtCore import QCoreApplication, Qt, QTimer, QProcess
from PyQt5.QtGui import QIcon, QCursor, QPixmap
import numpy as np
import matplotlib

matplotlib.use("Qt5Agg")  # noqa
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar,
)
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle
from matplotlib.lines import Line2D
import matplotlib.pyplot as plt

from homcloud.plot_PD_gui_ui import Ui_MainWindow
from homcloud.plot_PD import (
    add_arguments_for_auxinfo,
    add_arguments_for_zspec,
    AuxPlotInfo,
    ZSpec,
    PDColorHistogramPlotter,
    PDContourPlotter,
)
from homcloud.argparse_common import add_arguments_for_load_diagrams, add_arguments_for_histogram_rulers, parse_bool
import homcloud.utils as utils
import homcloud.resources  # noqa
from homcloud.version import __version__
from homcloud.histogram import Ruler, PDHistogram
import homcloud.pdgm as pdgm


class MainWindow(QMainWindow):
    def __init__(self, pd, phtrees_resolver, tmpdir, args, parent=None):
        super(MainWindow, self).__init__(parent)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        # initialize data
        self.pd = pd
        self.x_range = args.x_range or pd.minmax_of_birthdeath_time()
        self.y_range = args.y_range or self.x_range
        self.normalize_constant = args.normalize_constant
        self.aux_plot_info = AuxPlotInfo.from_args(args)

        # initialize event handlers for mouse events
        self.measure_distance_handler = MeasureDistanceHandler(self)
        self.rectangle_zoom_handler = RectangleZoomHandler(self)
        self.count_pairs_handler = CountPairsHandler(self)
        self.phtrees_handler = PHTreesHandler.create(phtrees_resolver, tmpdir, self)
        self.phtrees_range_query_handler = PHTreesRangeQueryHandler.create(phtrees_resolver, tmpdir, self)
        self.optimal_volume_handler = OptimalCycleHandler.create(
            pd, self, tmpdir, args.optimal_volume, shlex.split(args.optimal_volume_options)
        )
        self.current_canvas_ev_handler = NoOpMouseHandler()

        # setup UI part2
        self.setWindowTitle(args.title)
        self.figure = self.canvas = None
        self.add_matplotlib_canvas()
        self.ui.actionMeasure.setIcon(QIcon(":ui/icon_measure.png"))
        self.ui.actionRange.setIcon(QIcon(":ui/icon_range.png"))
        if not self.phtrees_handler.is_available():
            self.invalidate_phtrees_widgets()
        if not self.optimal_volume_handler.is_available():
            self.invalidate_optimal_volume_widgets()

        self.store_initial_widget_values(args)

        self.ax = self.background = None
        self.replot_history = []

        self.connect_signals()

        # setup unittest message mechanism
        self.test_mode = False
        self.test_message = None

    @staticmethod
    def quit():
        QCoreApplication.instance().quit()

    def add_matplotlib_canvas(self):
        self.figure = Figure()
        self.canvas = FigureCanvas(self.figure)
        self.ui.mplvl.addWidget(self.canvas)
        self.canvas.draw()
        self.addToolBar(NavigationToolbar(self.canvas, self, coordinates=True))

    def store_initial_widget_values(self, args):
        self.ui.edit_XMin.setText(str(self.x_range[0]))
        self.ui.edit_XMax.setText(str(self.x_range[1]))
        self.ui.edit_YMin.setText(str(self.y_range[0]))
        self.ui.edit_YMax.setText(str(self.y_range[1]))
        self.ui.edit_XBins.setText(str(args.xbins))
        self.ui.edit_YBins.setText(str(args.ybins or ""))
        self.ui.radioButton_linear.setChecked(True)
        if args.power is not None:
            self.ui.radioButton_power.setChecked(True)
            self.ui.edit_Exp.setText(args.power)
        self.ui.radioButton_log.setChecked(args.log)
        self.ui.radioButton_loglog.setChecked(args.loglog)
        self.ui.edit_VMax.setText(args.vmax)
        self.ui.edit_colormap.setText(args.colormap)

    def connect_signals(self):
        partial = functools.partial

        self.ui.actionQuit.triggered.connect(self.quit)
        self.ui.buttonReplot.clicked.connect(self.replot)
        self.ui.buttonUndo.clicked.connect(self.undo)

        self.ui.slider_XMin.valueChanged.connect(self.slidervaluechangeXmin)
        self.ui.slider_XMax.valueChanged.connect(self.slidervaluechangeXMax)
        self.ui.slider_YMin.valueChanged.connect(self.slidervaluechangeYmin)
        self.ui.slider_YMax.valueChanged.connect(self.slidervaluechangeYMax)
        self.ui.edit_XMin.editingFinished.connect(self.labelfinishedXmin)
        self.ui.edit_XMax.editingFinished.connect(self.labelfinishedXMax)
        self.ui.edit_YMin.editingFinished.connect(self.labelfinishedYmin)
        self.ui.edit_YMax.editingFinished.connect(self.labelfinishedYMax)

        self.ui.actionMeasure.triggered.connect(partial(self.activate_or_deactivate, self.measure_distance_handler))
        self.ui.actionRange.triggered.connect(partial(self.activate_or_deactivate, self.rectangle_zoom_handler))
        self.ui.actionCount.triggered.connect(partial(self.activate_or_deactivate, self.count_pairs_handler))
        self.ui.checkBox_query_phtree.clicked.connect(partial(self.activate_or_deactivate, self.phtrees_handler))
        self.ui.checkBox_show_ancestors.clicked.connect(self.draw_canvas)
        self.ui.checkBox_show_descendants.clicked.connect(self.draw_canvas)
        self.ui.button_show_descendants.clicked.connect(self.phtrees_handler.show_descendants_volumes)
        self.ui.checkBox_query_volumes_in_rectangle.clicked.connect(
            partial(self.activate_or_deactivate, self.phtrees_range_query_handler)
        )
        self.ui.checkBox_optimal_volume_on.clicked.connect(
            partial(self.activate_or_deactivate, self.optimal_volume_handler)
        )
        self.ui.checkBox_optimal_volume_show_children.clicked.connect(self.draw_canvas)

        self.canvas.mpl_connect("button_press_event", self.on_mouse_button_press)
        self.canvas.mpl_connect("button_release_event", self.on_mouse_button_release)
        self.canvas.mpl_connect("motion_notify_event", self.on_mouse_move)
        self.canvas.mpl_connect("axes_leave_event", self.on_mouse_leave)
        self.canvas.mpl_connect("draw_event", self.on_draw)
        self.canvas.mpl_connect("figure_enter_event", self.figure_enter)
        self.canvas.mpl_connect("figure_leave_event", self.figure_leave)

    def invalidate_phtrees_widgets(self):
        self.ui.checkBox_query_phtree.setEnabled(False)
        self.ui.button_show_descendants.setEnabled(False)
        self.ui.checkBox_show_ancestors.setEnabled(False)
        self.ui.checkBox_show_descendants.setEnabled(False)
        self.ui.checkBox_query_volumes_in_rectangle.setEnabled(False)
        self.ui.frame_phtree.setHidden(True)

    def invalidate_optimal_volume_widgets(self):
        self.ui.frame_optimal_volume.setHidden(True)

    def draw_canvas(self):
        try:
            inputbox_state = InputBoxState(self.ui)
        except ValueError as err:
            self.error_dialog("parameter convert error: {}".format(str(err)))
            return None

        histogram = PDHistogram(self.pd, *inputbox_state.xy_rulers(self.pd))
        histogram.multiply_histogram(1.0 / self.normalize_constant)

        if inputbox_state.apply_diffusion:
            histogram.apply_gaussian_filter(inputbox_state.diffusion_sigma)

        self.figure.clf()
        self.ax = self.figure.add_subplot(111)

        inputbox_state.plotter_class()(histogram, inputbox_state.zspec(), self.aux_plot_info).plot(
            self.figure, self.ax
        )

        self.phtrees_handler.draw_ancestors(self.ax)
        self.phtrees_handler.draw_descendants(self.ax)
        self.optimal_volume_handler.draw_children(self.ax)

        self.canvas.draw()
        self.save_backgroud()
        self.current_canvas_ev_handler.reset()
        return inputbox_state

    def save_backgroud(self):
        self.background = self.canvas.copy_from_bbox(self.ax.bbox)

    def restore_background(self):
        self.canvas.restore_region(self.background)

    def blit_canvas(self):
        self.canvas.blit(self.ax.bbox)

    def replot(self):
        inputbox_state = self.draw_canvas()
        if not inputbox_state:
            return
        self.replot_history.append(inputbox_state)
        self.set_buttonUndo()

    def set_buttonUndo(self):
        self.ui.buttonUndo.setEnabled(len(self.replot_history) >= 2)

    def xmin(self):
        return float(self.ui.edit_XMin.text())

    def xmax(self):
        return float(self.ui.edit_XMax.text())

    def ymin(self):
        return float(self.ui.edit_YMin.text())

    def ymax(self):
        return float(self.ui.edit_YMax.text())

    def undo(self):
        def adjust_sliders():
            self.labelfinishedXmin()
            self.labelfinishedXMax()
            self.labelfinishedYmin()
            self.labelfinishedYMax()

        self.replot_history.pop()
        self.replot_history[-1].restore(self.ui)
        adjust_sliders()
        self.draw_canvas()
        self.set_buttonUndo()

    def error_dialog(self, msg):
        if self.test_mode:
            self.test_message = msg
            return

        dialog = QMessageBox(self)
        dialog.setText(msg)
        dialog.exec_()

    def slidervaluechangeXmin(self):
        self.adjust_editline_by_slider(self.ui.slider_XMin, self.ui.edit_XMin, self.x_range)

    def slidervaluechangeXMax(self):
        self.adjust_editline_by_slider(self.ui.slider_XMax, self.ui.edit_XMax, self.x_range)

    def slidervaluechangeYmin(self):
        self.adjust_editline_by_slider(self.ui.slider_YMin, self.ui.edit_YMin, self.y_range)

    def slidervaluechangeYMax(self):
        self.adjust_editline_by_slider(self.ui.slider_YMax, self.ui.edit_YMax, self.y_range)

    @staticmethod
    def slider_to_value(val, range_):
        left, right = range_
        return left + val / 100.0 * (right - left)

    def adjust_editline_by_slider(self, slider, editline, range_):
        editline.setText(str(self.slider_to_value(slider.value(), range_)))

    def labelfinishedXmin(self):
        self.adjust_slider_by_edit(self.ui.slider_XMin, self.ui.edit_XMin, self.x_range)

    def labelfinishedXMax(self):
        self.adjust_slider_by_edit(self.ui.slider_XMax, self.ui.edit_XMax, self.x_range)

    def labelfinishedYmin(self):
        self.adjust_slider_by_edit(self.ui.slider_YMin, self.ui.edit_YMin, self.y_range)

    def labelfinishedYMax(self):
        self.adjust_slider_by_edit(self.ui.slider_YMax, self.ui.edit_YMax, self.y_range)

    def adjust_slider_by_edit(self, slider, editline, range_):
        try:
            value = float(editline.text())
        except ValueError:
            self.adjust_editline_by_slider(slider, editline, range_)
            self.error_dialog("Input data error")
            return
        slider.setSliderPosition(self.value_to_slider(value, range_))

    def set_slider_and_edit_by_range(self, slider, editline, value, range_):
        slider.setSliderPosition(self.value_to_slider(value, range_))
        editline.setText(str(value))

    @staticmethod
    def value_to_slider(val, range_):
        left, right = range_
        return int(100 * (val - left) / (right - left))

    def canvas_event_handlers(self):
        return [
            self.measure_distance_handler,
            self.rectangle_zoom_handler,
            self.count_pairs_handler,
            self.phtrees_handler,
            self.phtrees_range_query_handler,
            self.optimal_volume_handler,
        ]

    def activate_or_deactivate(self, handler, checked):
        if checked:
            handler.activate()
            for h in self.canvas_event_handlers():
                if h != handler:
                    h.activator_widget().setChecked(False)
        else:
            self.deactivate_mouse_handler()

    def on_mouse_button_press(self, event):
        self.current_canvas_ev_handler.on_button_press(event)

    def on_mouse_button_release(self, event):
        self.current_canvas_ev_handler.on_button_release(event)

    def on_mouse_move(self, event):
        self.current_canvas_ev_handler.on_mouse_move(event)

    def on_mouse_leave(self, event):
        self.current_canvas_ev_handler.on_mouse_leave(event)

    def deactivate_mouse_handler(self):
        self.current_canvas_ev_handler = NoOpMouseHandler()

    def on_draw(self, event):
        self.save_backgroud()

    def set_ui_ranges(self, x_range, y_range):
        self.set_slider_and_edit_by_range(self.ui.slider_XMin, self.ui.edit_XMin, x_range[0], self.x_range)
        self.set_slider_and_edit_by_range(self.ui.slider_XMax, self.ui.edit_XMax, x_range[1], self.x_range)
        self.set_slider_and_edit_by_range(self.ui.slider_YMin, self.ui.edit_YMin, y_range[0], self.y_range)
        self.set_slider_and_edit_by_range(self.ui.slider_YMax, self.ui.edit_YMax, y_range[1], self.y_range)

    def figure_enter(self, event):
        del event
        if self.ui.actionRange.isChecked():
            QApplication.setOverrideCursor(QCursor(Qt.CrossCursor))
        elif self.ui.actionMeasure.isChecked():
            open_hand_px = QPixmap(":ui/cursor_measure.png")
            open_hand_px.setMask(open_hand_px.mask())
            QApplication.setOverrideCursor(QCursor(open_hand_px))

    def figure_leave(self, event):
        del event
        if self.ui.actionRange.isChecked() or self.ui.actionMeasure.isChecked():
            QApplication.setOverrideCursor(QCursor(Qt.ArrowCursor))

    def get_test_message_and_clear(self):
        msg = self.test_message
        self.test_message = None
        return msg


class BaseMouseHandler(object):
    def is_valid_click_event(self, event):
        return (
            self.window.ax is not None
            and event.inaxes == self.window.ax
            and event.button == 1
            and event.xdata is not None
            and event.ydata is not None
        )

    def on_button_press(self, event):
        pass

    def on_mouse_move(self, event):
        pass

    def on_button_release(self, event):
        pass

    def on_mouse_leave(self, event):
        pass

    def reset(self):
        pass


class NullWidget(object):
    @staticmethod
    def setChecked(_):
        pass


class MouseDragHandler(BaseMouseHandler):
    def __init__(self, window):
        self.window = window
        self.start_xy = None

    def on_button_press(self, event):
        if not self.is_valid_click_event(event):
            return

        self.start_xy = (event.xdata, event.ydata)
        self.press()

    @staticmethod
    def press():
        pass

    def on_mouse_move(self, event):
        if not self.is_dragging(event):
            return

        mouse_xy = (event.xdata, event.ydata)
        self.draw(self.start_xy, mouse_xy)
        self.move(mouse_xy)

    @staticmethod
    def move(mouse_xy):
        pass

    def is_dragging(self, event):
        return event.inaxes == self.window.ax and self.start_xy

    def on_button_release(self, event):
        if not self.is_dragging(event):
            return

        start_xy = self.start_xy
        self.clear()
        self.release(start_xy, (event.xdata, event.ydata))

    @staticmethod
    def release(start_xy, end_xy):
        pass

    def on_mouse_leave(self, event):
        self.clear()

    def clear(self):
        self.start_xy = None
        self.window.restore_background()
        self.window.blit_canvas()

    def draw(self, start, end):
        self.window.restore_background()
        self.drawer.draw(start, end)
        self.window.blit_canvas()

    def reset(self):
        self.start_xy = None
        self.drawer.reset()

    class LineDrawer(object):
        def __init__(self, window):
            self.window = window
            self.line = None

        def reset(self):
            self.line = Line2D((0, 0), (0, 0), linewidth=1, color="black", linestyle="dotted")
            self.window.ax.add_line(self.line)

        def draw(self, start, end):
            self.line.set_xdata([start[0], end[0]])
            self.line.set_ydata([start[1], end[1]])
            self.window.ax.draw_artist(self.line)

    class RectangleDrawer(object):
        def __init__(self, window):
            self.window = window
            self.rect = None

        def reset(self):
            self.rect = Rectangle((0, 0), 0, 0, fill=False, linewidth="1.0", edgecolor="blue", linestyle="dotted")
            self.window.ax.add_patch(self.rect)

        def draw(self, start, end):
            xmin, xmax = minmax(start[0], end[0])
            ymin, ymax = minmax(start[1], end[1])
            self.rect.set_bounds(xmin, ymin, xmax - xmin, ymax - ymin)
            self.window.ax.draw_artist(self.rect)


class NoOpMouseHandler(BaseMouseHandler):
    pass


class MouseClickHandler(BaseMouseHandler):
    def __init__(self, window):
        self.window = window

    def on_button_press(self, event):
        if not self.is_valid_click_event(event):
            return
        self.click((event.xdata, event.ydata))

    def click(self, mouse_xy):
        pass


class MeasureDistanceHandler(object):
    def __init__(self, window):
        self.window = window

    def activate(self):
        self.window.current_canvas_ev_handler = self.MouseHandler(self.window)
        self.window.current_canvas_ev_handler.reset()

    def activator_widget(self):
        return self.window.ui.actionMeasure

    class MouseHandler(MouseDragHandler):
        def __init__(self, window):
            super().__init__(window)
            self.drawer = MouseDragHandler.LineDrawer(window)
            self.statusbar = window.ui.statusbar

        def show_distance(self, begin, end):
            def distance(u, v):
                return np.linalg.norm(np.array(u) - np.array(v))

            self.statusbar.showMessage("coord: {} - {},  distance: {}".format(begin, end, distance(begin, end)))

        def move(self, mouse_xy):
            self.show_distance(self.start_xy, mouse_xy)

        def press(self):
            self.show_distance(self.start_xy, self.start_xy)


class CountPairsHandler(object):
    def __init__(self, window):
        self.window = window

    def activate(self):
        self.window.current_canvas_ev_handler = self.MouseHandler(self.window)
        self.window.current_canvas_ev_handler.reset()

    def activator_widget(self):
        return self.window.ui.actionCount

    class MouseHandler(MouseDragHandler):
        def __init__(self, window):
            super().__init__(window)
            self.drawer = MouseDragHandler.RectangleDrawer(window)
            self.pd = window.pd
            self.statusbar = window.ui.statusbar
            self.timer = None

        def move(self, mouse_xy):
            self.stop_timer()
            self.timer = QTimer()
            self.timer.setSingleShot(True)
            self.timer.setInterval(100)
            self.timer.timeout.connect(functools.partial(self.show_pairs_in_rectangle, mouse_xy))
            self.timer.start()

        def show_pairs_in_rectangle(self, mouse_xy):
            if not self.start_xy:
                return

            num_pairs = self.pd.count_pairs_in_rectangle(self.start_xy[0], mouse_xy[0], self.start_xy[1], mouse_xy[1])
            self.statusbar.showMessage("coord: {} - {},  count: {}".format(self.start_xy, mouse_xy, num_pairs))

        def stop_timer(self):
            if self.timer:
                self.timer.stop()


class PHTreesHandler(object):
    def __init__(self, resolver, tmpdir, window):
        self.resolver = resolver
        self.leaf_for_query_ancestors = None
        self.root_for_query_descendants = None
        self.tmpdir = tmpdir
        self.window = window

    @staticmethod
    def is_available():
        return True

    def activate(self):
        self.window.current_canvas_ev_handler = self.MouseHandler(self)

    def activator_widget(self):
        return self.window.ui.checkBox_query_phtree

    def query(self, birth, death):
        self.leaf_for_query_ancestors = self.resolver.query_node(birth, death)
        self.root_for_query_descendants = self.resolver.query_node(birth, death)

    def show_descendants_volumes(self):
        if not self.root_for_query_descendants:
            return
        tmpfile = tmppath(self.tmpdir, ".vtk")
        self.root_for_query_descendants.draw_descendants_volumes(self.resolver.index_map.points, tmpfile, False, False)
        utils.invoke_paraview(tmpfile)

    def draw_ancestors(self, ax):
        if not self.window.ui.checkBox_show_ancestors.isChecked():
            return

        if not self.leaf_for_query_ancestors:
            return

        def visit_node(node):
            ax.scatter([node.birth_time()], [node.death_time()], c="blue", edgecolors="black")
            if node.parent:
                ax.plot(
                    [node.birth_time(), node.parent.birth_time()],
                    [node.death_time(), node.parent.death_time()],
                    c="black",
                )
                visit_node(node.parent)

        visit_node(self.leaf_for_query_ancestors)

    def draw_descendants(self, ax):
        if not self.window.ui.checkBox_show_descendants.isChecked():
            return

        if not self.root_for_query_descendants:
            return

        def visit_node(node):
            ax.scatter([node.birth_time()], [node.death_time()], c="red", edgecolors="black")

            for child in node.children:
                if not child.living():
                    continue
                ax.plot([node.birth_time(), child.birth_time()], [node.death_time(), child.death_time()], c="black")
                visit_node(child)

        visit_node(self.root_for_query_descendants)

    class MouseHandler(MouseClickHandler):
        def __init__(self, phtrees_handler):
            super().__init__(phtrees_handler.window)
            self.phtrees_handler = phtrees_handler

        def click(self, mouse_xy):
            self.phtrees_handler.query(*mouse_xy)
            self.window.draw_canvas()

    @staticmethod
    def create(resolver, tmpdir, window):
        if resolver is None:
            return PHTreesHandler.NoOp()
        else:
            return PHTreesHandler(resolver, tmpdir, window)

    class NoOp(object):
        @staticmethod
        def is_available():
            return False

        @staticmethod
        def activate():
            pass

        @staticmethod
        def activator_widget():
            return NullWidget()

        @staticmethod
        def show_descendants_volumes():
            pass

        @staticmethod
        def draw_ancestors(ax):
            del ax

        @staticmethod
        def draw_descendants(ax):
            del ax


class PHTreesRangeQueryHandler(object):
    def __init__(self, resolver, tmpdir, window):
        self.resolver = resolver
        self.tmpdir = tmpdir
        self.window = window

    def activate(self):
        self.window.current_canvas_ev_handler = self.MouseHandler(self)
        self.window.current_canvas_ev_handler.reset()

    def activator_widget(self):
        return self.window.ui.checkBox_query_volumes_in_rectangle

    def show_volumes_in_rectangle(self, v1, v2):
        xmin, xmax = minmax(v1[0], v2[0])
        ymin, ymax = minmax(v1[1], v2[1])
        nodes = self.resolver.query_nodes_in_rectangle(xmin, xmax, ymin, ymax)
        tmpfile = tmppath(self.tmpdir, ".vtk")
        self.resolver.draw_volumes_of_nodes(nodes, tmpfile, True, True)
        utils.invoke_paraview(tmpfile)

    class MouseHandler(MouseDragHandler):
        def __init__(self, phtrees_range_handler):
            super().__init__(phtrees_range_handler.window)
            self.drawer = MouseDragHandler.RectangleDrawer(phtrees_range_handler.window)
            self.phtrees_range_handler = phtrees_range_handler

        def release(self, start_xy, end_xy):
            self.phtrees_range_handler.show_volumes_in_rectangle(start_xy, end_xy)

    @staticmethod
    def create(resolver, tmpdir, ui):
        if resolver is None:
            return PHTreesRangeQueryHandler.NoOp()
        else:
            return PHTreesRangeQueryHandler(resolver, tmpdir, ui)

    class NoOp(object):
        @staticmethod
        def activate():
            pass

        @staticmethod
        def activator_widget():
            return NullWidget()


class RectangleZoomHandler(object):
    def __init__(self, window):
        self.window = window

    def activate(self):
        self.window.current_canvas_ev_handler = self.MouseHandler(self.window)
        self.window.current_canvas_ev_handler.reset()

    def activator_widget(self):
        return self.window.ui.actionRange

    class MouseHandler(MouseDragHandler):
        def __init__(self, window):
            super().__init__(window)
            self.drawer = MouseDragHandler.RectangleDrawer(window)

        def release(self, start_xy, end_xy):
            self.window.set_ui_ranges(minmax(start_xy[0], end_xy[0]), minmax(start_xy[1], end_xy[1]))
            self.window.replot()


class OptimalCycleHandler(object):
    def __init__(self, diagram_path, degree, window, tmpdir, options):
        self.diagram_path = diagram_path
        self.window = window
        self.degree = degree
        self.tmpdir = tmpdir
        self.options = options
        self.voc_info = None

    def activate(self):
        self.window.current_canvas_ev_handler = self.MouseHandler(self)

    def activator_widget(self):
        return self.window.ui.checkBox_optimal_volume_on

    def query(self, pos):
        vtkpath = tmppath(self.tmpdir, ".vtk")
        jsonpath = tmppath(self.tmpdir, ".json")
        options = self.options.copy()

        if self.window.ui.checkBox_optimal_volume_query_children.isChecked():
            options.extend(["-C", "on"])

        eigenvolume_th = self.window.ui.lineEdit_optimal_volume_stable_volume.text()
        if self.is_eigenvolume_none():
            pass
        elif self.window.ui.radioButton_eigenvolume_tightened.isChecked():
            options.extend(
                [
                    "--tightened-volume",
                    eigenvolume_th,
                    "--no-optimal-volume",
                    "yes",
                    "--show-optimal-volume",
                    "no",
                ]
            )
        elif self.window.ui.radioButton_eigenvolume_owned.isChecked():
            options.extend(
                [
                    "--owned-volume",
                    eigenvolume_th,
                    "--show-optimal-volume",
                    "no",
                ]
            )
        elif self.window.ui.radioButton_eigenvolume_tightened_sub.isChecked():
            options.extend(
                [
                    "--tightened-subvolume",
                    eigenvolume_th,
                    "--show-optimal-volume",
                    "no",
                ]
            )

        process = QProcess()
        process.setProcessChannelMode(QProcess.ForwardedChannels)
        dialog = QProgressDialog("Computing volume optimal cycle...", "Cancal", 0, 0)
        process.finished.connect(dialog.reset)
        dialog.canceled.connect(process.kill)

        process.start(
            sys.executable,
            [
                "-m",
                "homcloud.optimal_volume",
                "-d",
                str(self.degree),
                "-x",
                str(pos[0]),
                "-y",
                str(pos[1]),
                "-v",
                vtkpath,
                "-n",
                self.window.ui.lineEdit_optimal_volume_retry.text(),
                "--cutoff-radius",
                self.window.ui.lineEdit_optimal_volume_cutoff_radius.text(),
                "-j",
                jsonpath,
                self.diagram_path,
            ]
            + options,
        )
        if not process.waitForStarted():
            self.show_optimal_volume_error(-1)
            return
        dialog.exec()
        if process.waitForFinished():
            return
        elif process.exitCode() == 0:
            if self.is_eigenvolume_none():
                self.load_info_from_jsonfile(jsonpath)
            self.window.draw_canvas()
            if self.window.ui.checkBox_optimal_volume_show_volume.isChecked():
                utils.invoke_paraview(vtkpath)
        else:
            self.show_optimal_volume_error(process.exitCode())
        # Here, the above code is difficult to understand, hence we explain as follow:
        #
        # First, by `procees.start`, the optimal cycle program starts in prallel with
        # the GUI process.
        # Next, call QProcess.waitForStarted to check the program is successfully started
        # Next, by `dialog.exec`, the dialog is invoked.
        # This dialog is modal, we will wait for finishing the process or
        # we click cancel button.
        #   * Then, if the program finishes, then dialog is closed from
        #     the QProcess#finished signal
        #     * In this case, process.waitForFinished returns false since the process is
        #       already stopped
        #
        #   * If the cancel button is clicked, the process is killed from
        #     QProgressDialog#canceled signal
        #     * In this case, process.waitForFinished returns true or false
        #       * It return true when kill signal is sent but the process is not yet killed
        #         because of the timing problem
        #       * Otherwise, it returns false
        # In both case, after `dialog.exec()`, the process is "almost" stopped.

    def show_optimal_volume_error(self, exitcode):
        if exitcode == 10:
            msg = "Death simplex out of cutoff ball"
        elif exitcode == 2:
            msg = "Infeasible (probably cutoff radius is too small)"
        elif exitcode == -1:
            msg = "Process invoke error"
        else:
            msg = "Unknown code: {}".format(exitcode)

        self.window.error_dialog(msg)

    def load_info_from_jsonfile(self, path):
        with open(path) as f:
            self.voc_info = json.load(f)["result"]

    def is_eigenvolume_none(self):
        return self.window.ui.radioButton_eigenvolume_none.isChecked()

    def draw_children(self, ax):
        if not self.voc_info:
            return
        if not self.window.ui.checkBox_optimal_volume_show_children.isChecked():
            return

        ax.scatter([self.voc_info[0]["birth-time"]], [self.voc_info[0]["death-time"]], c="green", edgecolor="black")
        for child in self.voc_info[0]["children"]:
            ax.scatter([child["birth-time"]], [child["death-time"]], c="red", edgecolor="black")

    @staticmethod
    def is_available():
        return True

    class MouseHandler(MouseClickHandler):
        def __init__(self, optimal_volume_handler):
            super().__init__(optimal_volume_handler.window)
            self.optimal_volume_handler = optimal_volume_handler

        def click(self, mouse_xy):
            self.optimal_volume_handler.query(mouse_xy)

    @staticmethod
    def create(pd, window, tmpdir, activate, options):
        def optimal_volume_available():
            return pd.filtration_type == "alpha" and pd.path

        if optimal_volume_available() and activate:
            return OptimalCycleHandler(pd.path, pd.degree, window, tmpdir, options)
        else:
            return OptimalCycleHandler.NoOp()

    class NoOp(object):
        @staticmethod
        def clicked(pos):
            del pos

        @staticmethod
        def is_available():
            return False

        @staticmethod
        def activator_widget():
            return NullWidget()

        @staticmethod
        def draw_children(ax):
            return

    class Dialog(QDialog):
        pass


class InputBoxState(object):
    def __init__(self, ui):
        self.normalplot = ui.radioButton_normalplot.isChecked()
        self.contourplot = ui.radioButton_contourplot.isChecked()
        self.linear_z = ui.radioButton_linear.isChecked()
        self.log_z = ui.radioButton_log.isChecked()
        self.loglog_z = ui.radioButton_loglog.isChecked()
        self.power_z = ui.radioButton_power.isChecked()
        self.power_z_exp_text = ui.edit_Exp.text()
        self.z_max_text = ui.edit_VMax.text()
        self.apply_diffusion = ui.checkBox_Diffusing.isChecked()
        self.diffusing_sigma_text = ui.edit_DiffusingSize.text()
        self.xmin_text = ui.edit_XMin.text()
        self.xmax_text = ui.edit_XMax.text()
        self.xrange_eq_yrange = ui.checkBox_yrange.isChecked()
        self.ymin_text = ui.edit_YMin.text()
        self.ymax_text = ui.edit_YMax.text()
        self.xbins_text = ui.edit_XBins.text()
        self.ybins_text = ui.edit_YBins.text()
        self.colormap = ui.edit_colormap.text()
        self.validate_colormap()

        if self.power_z:
            self.power_z_exp = float(self.power_z_exp_text)

        self.z_max = float(self.z_max_text) if self.z_max_text else None

        if self.apply_diffusion:
            self.diffusion_sigma = float(self.diffusing_sigma_text)
        else:
            self.diffusion_sigma = None

        self.xmin = float(self.xmin_text)
        self.xmax = float(self.xmax_text)
        self.ymin = self.xmin if self.xrange_eq_yrange else float(self.ymin_text)
        self.ymax = self.xmax if self.xrange_eq_yrange else float(self.ymax_text)
        self.xbins = int(self.xbins_text)
        self.ybins = int(self.ybins_text) if self.ybins_text else None

    def validate_colormap(self):
        if self.colormap:
            plt.get_cmap(self.colormap)

    def xy_rulers(self, diagram):
        return Ruler.create_xy_rulers((self.xmin, self.xmax), self.xbins, (self.ymin, self.ymax), self.ybins, diagram)

    def zspec(self):
        colormap = plt.get_cmap(self.colormap) if self.colormap else None

        if self.linear_z:
            return ZSpec.Linear(self.z_max, None, colormap)
        if self.log_z and self.apply_diffusion:
            return ZSpec.LogWithDiffusion(self.z_max, None, colormap)
        if self.log_z:
            return ZSpec.Log(self.z_max, None, colormap)
        if self.loglog_z:
            return ZSpec.LogLog(self.z_max, None, colormap)
        if self.power_z:
            return ZSpec.Power(self.power_z_exp, self.z_max, None, colormap)
        raise ValueError("zspec radiobutton error")

    def plotter_class(self):
        if self.normalplot:
            return PDColorHistogramPlotter
        if self.contourplot:
            return PDContourPlotter
        raise ValueError("plot style radiobutton error")

    def restore(self, ui):
        ui.radioButton_normalplot.setChecked(self.normalplot)
        ui.radioButton_contourplot.setChecked(self.contourplot)
        ui.radioButton_linear.setChecked(self.linear_z)
        ui.radioButton_log.setChecked(self.log_z)
        ui.radioButton_loglog.setChecked(self.loglog_z)
        ui.radioButton_power.setChecked(self.power_z)
        ui.edit_Exp.setText(self.power_z_exp_text)
        ui.edit_VMax.setText(self.z_max_text)
        ui.edit_colormap.setText(self.colormap)
        ui.checkBox_Diffusing.setChecked(self.apply_diffusion)
        ui.edit_DiffusingSize.setText(self.diffusing_sigma_text)
        ui.edit_XMin.setText(self.xmin_text)
        ui.edit_XMax.setText(self.xmax_text)
        ui.checkBox_yrange.setChecked(self.xrange_eq_yrange)
        ui.edit_YMin.setText(self.ymin_text)
        ui.edit_YMax.setText(self.ymax_text)
        ui.edit_XBins.setText(self.xbins_text)


def minmax(x, y):
    return min(x, y), max(x, y)


def tmppath(tmpdir, suffix):
    return os.path.join(tmpdir, str(uuid.uuid4()) + suffix)


def argument_parser():
    p = argparse.ArgumentParser(description="View PD interactively", conflict_handler="resolve")
    p.add_argument("-V", "--version", action="version", version=__version__)
    add_arguments_for_load_diagrams(p)
    p.add_argument("--power", default=None, help="colorbar scale: x^POWER for each value x")
    p.add_argument("-l", "--log", action="store_true", default=False, help="colorbar scale: log(x+1) for each value x")
    p.add_argument("--loglog", action="store_true", default=False, help="colorbar scale: log(log(x+1)+1)")
    p.add_argument("-m", "--vmax", metavar="MAX", default="", help="Maximum of colorbar (default: autoscale)")
    p.add_argument("-c", "--colormap", default="", help="matplotlib colormap name")
    add_arguments_for_histogram_rulers(p)
    add_arguments_for_auxinfo(p)
    p.add_argument(
        "-n", "--normalize-constant", type=float, default=1.0, help="normalize constant to histogram height"
    )
    p.add_argument("-p", "--ph-trees", help="ph trees file (.pht)")
    p.add_argument(
        "--optimal-volume", "--optimal-cycle", default=False, type=parse_bool, help="activate optimal volume (on/off)"
    )
    p.add_argument("--optimal-volume-options", default="", help="options for optimal_volume module")
    p.add_argument("input", nargs="+", help="input file")
    return p


def main():
    app = QApplication(sys.argv)
    args = argument_parser().parse_args(app.arguments()[1:])
    if args.title is None:
        args.title = args.input[0]

    pd = pdgm.load_merged_diagrams(args.input, args.type, args.degree)

    if args.ph_trees:
        with open(args.ph_trees, "rb") as f:
            phtrees_resolver = full_ph_tree.TreeResolver.load(f)
    else:
        phtrees_resolver = None

    with tempfile.TemporaryDirectory() as tmpdir:
        main_window = MainWindow(pd, phtrees_resolver, tmpdir, args)
        app.lastWindowClosed.connect(MainWindow.quit)

        main_window.replot()
        main_window.show()

        sys.exit(app.exec_())


if __name__ == "__main__":
    main()



import re
import math
import argparse

import msgpack

from homcloud.argparse_common import parse_bool
from homcloud.version import __version__
from homcloud.license import add_argument_for_license
from homcloud.pdgm_format import PDGMWriter, SimpleChunk, BoundaryMapChunk
import homcloud.phat_ext as phat


def main(args=None):
    args = args or argument_parser().parse_args()
    check_output_path(args.output)
    with open(args.input) as f:
        filt = AbstractFiltration.load(f, args.save_boundary_map)
    with open(args.output, "wb") as f:
        filt.compute_pdgm(f)


def check_output_path(path):
    if path.endswith(".idiagram") or path.endswith(".diagram"):
        raise RuntimeError(
            ".idiagram and .diagram format is not available on this program. " "Please use .pdgm file instead"
        )


def argument_parser():
    p = argparse.ArgumentParser(description="Convert a description of boundary map to a PD")
    p.add_argument("-V", "--version", action="version", version=__version__)
    p.add_argument(
        "-M", "--save-boundary-map", default=True, type=parse_bool, help="save boundary map into idiagram file"
    )
    add_argument_for_license(p)
    p.add_argument("input", help="input file name")
    p.add_argument("output", help="output file name")
    return p


class AbstractFiltration(object):
    def __init__(self, boundary_map, dim, levels, symbols, save_boundary_map):
        self.boundary_map = boundary_map
        self.dim = dim
        self.index_to_level = levels
        self.index_to_symbol = symbols
        self.save_boundary_map = save_boundary_map

    @staticmethod
    def load(f, save_boundary_map):
        loader = AbstractFiltrationLoader(f)
        loader.load()
        return loader.filtration(save_boundary_map)

    def build_phat_matrix(self):
        matrix = phat.Matrix(len(self.boundary_map), "none")
        for n, single in enumerate(self.boundary_map):
            dim = single[0]
            boundary = []
            for idx, coef in zip(single[1], single[2]):
                if coef % 2 != 0:
                    boundary.append(idx)
            matrix.set_dim_col(n, dim, sorted(boundary))

        return matrix

    def compute_pdgm(self, f, algorithm=None, output_suppl_info=True):
        matrix = self.build_phat_matrix()
        matrix.reduce(algorithm)

        writer = PDGMWriter(f, "abstract", self.dim)
        writer.save_pairs(matrix.birth_death_pairs(), self.index_to_level, output_suppl_info)
        if output_suppl_info:
            writer.append_simple_chunk("index_to_level", self.index_to_level)
            writer.append_simple_chunk("index_to_symbol", self.index_to_symbol)

        if self.save_boundary_map:
            writer.append_chunk(BoundaryMapChunk("abstract", self.boundary_map))

        writer.write()

    @staticmethod
    def favorite_algorithm():
        return "phat-twist"


class ParseError(RuntimeError):
    pass


class AbstractFiltrationLoader(object):
    def __init__(self, io):
        self.io = io
        self.autoid = False
        self.autosymbol = True
        self.nextid = 0
        self.lasttime = -math.inf
        self.boundary_information = []

    def load(self):
        for line in self.io:
            line = line.strip()
            if line == "" or self.iscomment(line):
                continue
            if self.is_optionline(line):
                self.parse_option(line)
                continue
            self.boundary_information.append(self.parse_line(line))
            self.nextid += 1

    def filtration(self, save_boundary_map):
        symbols = []
        levels = []
        boundary_map = []
        dims = []
        for id, symbol, dim, time, indices, coefs in self.boundary_information:
            symbols.append(symbol)
            levels.append(time)
            dims.append(dim)
            boundary_map.append([dim, indices, coefs])

        return AbstractFiltration(boundary_map, max(dims), levels, symbols, save_boundary_map)

    OPTIONS = [re.compile(r"autoid\s*:"), re.compile(r"autosymbol\s*:")]

    def is_optionline(self, line):
        return any(option.match(line) for option in self.OPTIONS)

    def parse_option(self, line):
        if self.boundary_information:
            raise ParseError("Options must be above boundary information")
        command, _, args = line.lower().partition(":")
        args = args.strip()
        if command == "autoid":
            self.autoid = parse_bool(args)
        if command == "autosymbol":
            self.autosymbol = parse_bool(args)

    @staticmethod
    def iscomment(line):
        return line.startswith("#")

    def parse_line(self, line):
        def check_id(id):
            if id != self.nextid:
                raise ParseError("id must be incremental integers from 0")

        def check_symbol(symbol):
            if not re.match(r"[a-zA-Z0-9_]+\Z", symbol):
                raise ParseError("Invalid cell symbol {}".format(symbol))

        def check_time(time):
            if time < self.lasttime:
                raise ParseError("time must be monotone")
            self.lasttime = time

        if line.count("=") != 1:
            raise ParseError('"=" symbol is required')

        leftstr, rightstr = line.split("=")

        indices, coefs = self.parse_boundary(rightstr)

        left = re.split(r"\s+", leftstr.strip())
        id = self.nextid if self.autoid else int(left.pop(0), 10)
        check_id(id)
        symbol = str(id) if self.autosymbol else left.pop(0)
        check_symbol(symbol)
        dim = int(left.pop(0), 10)
        time = float(left.pop(0))
        check_time(time)
        if left:
            raise ParseError('Too much elements in a line: "{}"'.format(line))

        return id, symbol, dim, time, indices, coefs

    def parse_boundary(self, s):
        def parse_numbers(string):
            if string:
                return [int(n, 10) for n in re.split(r"\s+", string.strip())]
            else:
                return []

        if s.count(":") != 1:
            raise ParseError('":" symbol is required')
        indices_str, coef_str = re.split(r":", s.strip())
        return parse_numbers(indices_str), parse_numbers(coef_str)


if __name__ == "__main__":
    main()



import numpy as np
from cached_property import cached_property

import homcloud.maximal_0_component as maximal_0_component


class PH0Components(object):
    def __init__(self, pair, epsilon):
        self.pair = pair
        self.graph = maximal_0_component.Graph(pair.diagram.pd, pair.birth_index, pair.death_index, epsilon)

    @property
    def _symbol_resolver(self):
        return self.pair.diagram.pd.alpha_symbol_resolver

    @property
    def _coord_resolver(self):
        return self.pair.diagram.pd.alpha_coord_resolver

    @property
    def birth_component(self):
        return np.array([self._coord_resolver.vertices[i] for i in self.graph.birth_component])

    @property
    def elder_component(self):
        return np.array([self._coord_resolver.vertices[i] for i in self.graph.elder_component])

    @property
    def birth_component_symbols(self):
        return [self._symbol_resolver.vertices[i] for i in self.graph.birth_component]

    @property
    def elder_component_symbols(self):
        return [self._symbol_resolver.vertices[i] for i in self.graph.elder_component]



import argparse
import math

import numpy as np
import scipy.ndimage
import scipy.misc
from scipy import stats
import matplotlib.pyplot as plt
from sklearn import mixture, cluster
from .mixed_gaussian import WeightedNormal


class RangeAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        begin, end = values.split(":")
        setattr(namespace, "pixelvalue_range", (float(begin), float(end)))


def load_pixels(path):
    image = scipy.ndimage.imread(path, flatten=True)
    if np.max(image) > 255:
        image //= 256

    return image


def plot_gaussians(gaussians, scale):
    for gaussian in gaussians:
        xs = np.linspace(0, 255, 256)
        ys = scale * gaussian.density(xs)
        plt.plot(xs, ys)
        plt.axvline(gaussian.mean, color="r")
        plt.axvline(gaussian.mean + gaussian.sd / 2, color="r")
        plt.axvline(gaussian.mean - gaussian.sd / 2, color="r")


def gmm_plot(pixels, n_components):
    gmm = mixture.GMM(n_components=n_components)
    gmm.fit(pixels)
    plot_gaussians(WeightedNormal.from_gmm(gmm), pixels.size)


class Cluster(object):
    def __init__(self, center, data):
        self.center = center
        self.data = data.flatten()
        (self.size,) = self.data.shape
        self.var = np.cov(self.data)
        self.sd = np.sqrt(self.var)

    def loglikelyhood(self):
        return np.sum(stats.norm.logpdf(self.data, self.center, self.sd))

    def f(self):
        return self.size * math.log(self.var)


def kmeans_bic(kmeans, pixels, n_components):
    clusters = [Cluster(kmeans.cluster_centers_[k, 0], pixels[kmeans.labels_ == k, :]) for k in range(0, n_components)]
    llh = sum([cl.loglikelyhood() for cl in clusters])
    return -2 * llh + 2 * n_components * math.log(pixels.size)


def kmeans_sic(kmeans, pixels, n_components):
    clusters = [Cluster(kmeans.cluster_centers_[k, 0], pixels[kmeans.labels_ == k, :]) for k in range(0, n_components)]
    return sum([cl.f() for cl in clusters]) + 2 * n_components * math.log(pixels.size)


def kmeans_plot(pixels, n_components):
    kmeans = cluster.KMeans(n_components)
    kmeans.fit(pixels)

    for k in range(0, n_components):
        cl = pixels[kmeans.labels_ == k, :]
        center = kmeans.cluster_centers_[k]
        boundary = np.max(cl)
        plt.axvline(boundary, color="g")
        plt.axvline(center, color="r")
        sd = math.sqrt(np.cov(cl.T))
        plt.axvline(center - sd / 2, color="r")
        plt.axvline(center + sd / 2, color="r")

    # print(kmeans_bic(kmeans, pixels, n_components))
    # print(kmeans_sic(kmeans, pixels, n_components))


def dpgmm_plot(pixels, n_components):
    dpgmm = mixture.DPGMM(n_components=n_components)
    dpgmm.fit(pixels)
    plot_gaussians(WeightedNormal.from_dpgmm(dpgmm), pixels.size)


def vbgmm_plot(pixels, n_components):
    vbgmm = mixture.VBGMM(n_components=n_components, alpha=1000)
    vbgmm.fit(pixels)
    print(vbgmm.n_components)
    plot_gaussians(WeightedNormal.from_dpgmm(vbgmm), pixels.size)


def dbscan_plot(pixels):
    dbscan = cluster.DBSCAN(eps=50)
    dbscan.fit(pixels)
    print(dbscan.core_sample_indices_)


def filter_by_range(pixels, minvalue, maxvalue):
    return pixels[(minvalue <= pixels) & (pixels <= maxvalue)]



# pylint: disable=R0903
import argparse
import sys
import copy
import re

import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.cm as cm
from matplotlib.lines import Line2D
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np

import homcloud.pdgm as pdgm
from homcloud.histogram import PDHistogram, Ruler
from homcloud.norms_for_matplotlib import LogPlusOneNorm, LogLogNorm, MidpointNormalize
from homcloud.version import __version__
from homcloud.argparse_common import add_arguments_for_load_diagrams, add_arguments_for_histogram_rulers


def main(args=None):
    """The main routine"""
    args = args or argument_parser().parse_args()
    pd = pdgm.load_merged_diagrams(args.input, args.type, args.degree)
    xy_rulers = Ruler.create_xy_rulers(args.x_range, args.xbins, args.y_range, args.ybins, pd)
    histogram = PDHistogram(pd, *xy_rulers)
    histogram.multiply_histogram(1.0 / args.normalize_constant)
    if args.diffuse_pairs:
        histogram.apply_gaussian_filter(args.diffuse_pairs)
    zspec = ZSpec.create_from_args(args)
    if args.title is None:
        args.title = args.input[0]
    aux_info = AuxPlotInfo.from_args(args)
    plotter = PDPlotter.find_plotter(args.style)(histogram, zspec, aux_info)

    fig, ax = plt.subplots()
    plotter.plot(fig, ax)

    if args.marker:
        MarkerDrawer.load_from_file(args.marker).draw(ax)

    if args.tight_layout or aux_info.require_tight_layout():
        plt.tight_layout()

    if args.output:
        plt.savefig(args.output, dpi=args.dpi)
    else:
        plt.show()

    plt.close(fig)


POWER_NORM_IS_AVAILABLE = hasattr(colors, "PowerNorm")


def argument_parser():
    parser = argparse.ArgumentParser(description="Plot an output by dipha")
    parser.add_argument("-V", "--version", action="version", version=__version__)
    add_arguments_for_load_diagrams(parser)
    add_arguments_for_zspec(parser)
    parser.add_argument(
        "-s", "--style", default="colorhistogram", help="plotting style (colorhistogram(default), contour)"
    )
    add_arguments_for_auxinfo(parser)
    parser.add_argument(
        "-D",
        "--diffuse-pairs",
        metavar="SCATTERING_SIZE",
        type=float,
        default=None,
        help="Diffuse pairs using gaussian distribution of SD=SIGMA",
    )
    parser.add_argument("-o", "--output", help="output file")
    add_arguments_for_histogram_rulers(parser)
    parser.add_argument(
        "-n", "--normalize-constant", type=float, default=1.0, help="normalize constant to histogram height"
    )
    parser.add_argument("-M", "--marker", help="marker file")
    parser.add_argument(
        "--dpi", type=int, default=None, help="output DPI (used with -o option, default is savefig.dpi for matplotlib)"
    )
    parser.add_argument(
        "--tight-layout", action="store_true", default=False, help="use tight layout (adjusting layout)"
    )
    parser.add_argument("input", metavar="INPUT", nargs="+", help="Input file path")
    return parser


def add_arguments_for_zspec(parser):
    if POWER_NORM_IS_AVAILABLE:
        parser.add_argument(
            "-p", "--power", metavar="POWER", type=float, default=None, help="Output x^POWER for each value x"
        )
    parser.add_argument("-l", "--log", action="store_true", default=False, help="Output log(x+1) for each value x")
    parser.add_argument("--loglog", action="store_true", default=False, help="Output log(log(x+1)+1)")
    parser.add_argument("--linear-midpoint", type=float, help="linear with midpoint")
    parser.add_argument(
        "-m", "--vmax", metavar="MAX", type=float, default=None, help="Maximum of colorbar (default: autoscale)"
    )
    parser.add_argument("--vmin", type=float, default=None, help="Minimum of colorbar")
    parser.add_argument("-c", "--colormap", default=None, help="matplotlib colormap name")


def add_arguments_for_auxinfo(parser):
    parser.add_argument("-t", "--title", help="title string")
    parser.add_argument("-U", "--unit-name", help="The unit name of birth and death times")
    parser.add_argument("--font-size", help="font size (default: 12)")
    parser.add_argument("--aspect", default="auto", help="histogram aspect (default: auto)")
    parser.add_argument(
        "--plot-essential", action="store_true", help="whether to plot essential values (default: False)"
    )


class AuxPlotInfo(object):
    def __init__(self, title, unit_name, font_size=None, aspect="auto", plot_ess=False):
        self.title = title
        self.unit_name = unit_name
        self.font_size = font_size
        self.aspect = aspect
        self.plot_ess = plot_ess

    @staticmethod
    def from_args(args):
        return AuxPlotInfo(args.title, args.unit_name, args.font_size, args.aspect, args.plot_essential)

    def require_tight_layout(self):
        return self.font_size is not None


class PDPlotter(object):
    """Persistence diagram plotter"""

    def __init__(self, histogram, zspec, aux_info):
        self.histogram = histogram
        self.zspec = zspec
        self.aux_info = aux_info
        self.qmesh = None
        self.qmesh_ess = None
        self.set_colorbar_range()

    def plot(self, fig, ax):
        """Plot the diagram to (fix, ax).

        Args:
        fig -- matplotlib.figures.Figure
        ax -- matplotlib.figure.Axes object related to fig
        """
        self.set_plot_range(ax)
        self.draw_labels(ax)
        self.set_title(ax)
        self.draw_diagonal(ax)
        self.set_tick_font_size(ax)
        self.plot_main_box(ax)
        self.draw_grid(ax)
        self.set_aspect(ax)
        self.divider(fig, ax)

    def set_colorbar_range(self):
        if self.zspec.vmax is None:
            self.zspec.vmax = self.histogram.maxvalue()

    def set_plot_range(self, ax):
        ax.set_xlim(*self.histogram.x_range())
        ax.set_ylim(*self.histogram.y_range())

    def draw_labels(self, ax):
        ax.set_xlabel(self.label_text("Birth"), fontsize=self.aux_info.font_size)
        ax.set_ylabel(self.label_text("Death"), fontsize=self.aux_info.font_size)

    def set_title(self, ax):
        if not self.aux_info.plot_ess:
            ax.set_title(self.aux_info.title, fontsize=self.aux_info.font_size)

    def label_text(self, key):
        """Format label text and return the text."""
        if self.aux_info.unit_name is None:
            return key
        else:
            return "{}[{}]".format(key, self.aux_info.unit_name)

    def draw_diagonal(self, ax):
        x_range = self.histogram.x_range()
        ax.add_line(Line2D(x_range, x_range, linewidth=1, color="black"))

    def set_tick_font_size(self, ax):
        ax.tick_params(labelsize=self.aux_info.font_size)

    def set_aspect(self, ax):
        if self.aux_info.aspect in ["auto", "equal"]:
            ax.set_aspect(self.aux_info.aspect)

    def divider(self, fig, ax):
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", "5%", pad="3%")
        colorbar = fig.colorbar(self.qmesh, cax=cax)
        colorbar.ax.tick_params(labelsize=self.aux_info.font_size)
        colorbar.set_label("Value")

        if self.aux_info.plot_ess:
            if self.histogram.sign_flipped:
                ax_ess = divider.append_axes("bottom", "5%", pad=0)
                ax.tick_params(bottom=False, labelbottom=False)
                ax_ess.set_ylabel(self.label_text("-∞"), fontsize=self.aux_info.font_size)
                ax_ess.set_xlabel(self.label_text("Birth"), fontsize=self.aux_info.font_size)
                ax.set_title(self.aux_info.title, fontsize=self.aux_info.font_size)
            else:
                ax_ess = divider.append_axes("top", "5%", pad=0)
                ax_ess.set_ylabel(self.label_text("∞"), fontsize=self.aux_info.font_size)
                ax_ess.set_title(self.aux_info.title, fontsize=self.aux_info.font_size)

            ax_ess.set_xlim(*self.histogram.x_range())
            ax_ess.tick_params(bottom=False, left=False, labelbottom=False, labelleft=False)
            ax_ess.axhline(y=1.5, c="k", lw="0.3")

            ess_yedges = np.linspace(0.0, 3.0, num=4)
            zeros = np.zeros((3, self.histogram.x_bins()))
            zeros[1] = self.histogram.ess_values
            self.histogram.ess_values_as_float = zeros.astype(float)
            self.qmesh_ess = ax_ess.pcolormesh(
                self.histogram.xedges,
                ess_yedges,
                self.histogram.ess_values_as_float,
                norm=self.zspec.norm(),
                cmap=self.zspec.colormap(),
            )

            ax_ess.grid(axis="x")

    @staticmethod
    def find_plotter(style):
        """Return appropriate plotter class from *style* argument."""
        if style == "colorhistogram":
            return PDColorHistogramPlotter
        elif style == "contour":
            return PDContourPlotter
        else:
            raise ValueError("Unknown plot style: %s" % style)

    @staticmethod
    def draw_grid(ax):
        ax.grid()


class PDColorHistogramPlotter(PDPlotter):
    def plot_main_box(self, ax):
        self.qmesh = ax.pcolormesh(
            self.histogram.xedges,
            self.histogram.yedges,
            self.histogram.values_as_float,
            norm=self.zspec.norm(),
            cmap=self.zspec.colormap(),
        )


class PDContourPlotter(PDPlotter):
    def __init__(self, histogram, zspec, aux_info, levels=None):
        super().__init__(histogram, zspec, aux_info)
        self.levels = levels

    def plot_main_box(self, ax):
        self.qmesh = ax.contour(
            self.histogram.values_as_float,
            norm=self.zspec.norm(),
            cmap=cm.rainbow,
            extent=self.histogram.xy_extent(),
            levels=self.contour_levels(),
        )

    def contour_levels(self):
        """
        Return list of levels (floats) for matplotlib.axis.Axes.contour.
        The levels is appropriately computed by zspec.
        """
        if self.levels is not None:
            return self.levels
        norm = self.zspec.norm()
        norm.autoscale_None(self.histogram.values_as_float)
        return norm.inverse(np.linspace(0.01, 1.0, 6))


class ZSpec(object):
    """
    Classes under ZSpec hold the information about the visualization of
    z-axis (the number of points in each bin) of the histogram.
    """

    @staticmethod
    def create_from_args(args):
        """Create a zspec object selected by args."""
        colormap = plt.get_cmap(args.colormap) if args.colormap else None
        if args.power:
            return ZSpec.Power(args.power, args.vmax, args.vmin, colormap)
        if args.log and args.diffuse_pairs:
            return ZSpec.LogWithDiffusion(args.vmax, args.vmin, colormap)
        if args.log:
            return ZSpec.Log(args.vmax, args.vmin, colormap)
        if args.loglog:
            return ZSpec.LogLog(args.vmax, args.vmin, colormap)
        if args.linear_midpoint is not None:
            return ZSpec.LinearMidPoint(args.linear_midpoint, args.vmax, args.vmin, colormap)
        return ZSpec.Linear(args.vmax, args.vmin, colormap)

    class Base(object):
        def __init__(self, vmax=None, vmin=None, colormap=None):
            self.vmin = vmin
            self.vmax = vmax
            self.cmap = colormap

        @staticmethod
        def default_colormap():
            return cm.OrRd

        def colormap(self):
            return self.cmap or self.default_colormap()

    class Log(Base):
        def norm(self):
            return colors.LogNorm(vmax=self.vmax, vmin=self.vmin)

        @staticmethod
        def default_colormap():
            cmap = copy.copy(cm.rainbow)
            cmap.set_under("white")
            return cmap

    class Power(Base):
        def __init__(self, power, vmax=None, vmin=None, colormap=None):
            self.power = power
            super().__init__(vmax, vmin, colormap)

        def norm(self):
            return colors.PowerNorm(self.power, vmax=self.vmax, vmin=self.vmin)

    class LogWithDiffusion(Base):
        def norm(self):
            return LogPlusOneNorm(vmax=self.vmax, vmin=self.vmin)

    class LogLog(Base):
        def norm(self):
            return LogLogNorm(vmax=self.vmax, vmin=self.vmin)

    class Linear(Base):
        def norm(self):
            return colors.Normalize(vmax=self.vmax, vmin=self.vmin)

    class LinearMidPoint(Base):
        def __init__(self, midpoint, vmax=None, vmin=None, colormap=None):
            self.midpoint = midpoint
            super().__init__(vmax, vmin, colormap)

        def norm(self):
            return MidpointNormalize(vmax=self.vmax, vmin=self.vmin, midpoint=self.midpoint)

        @staticmethod
        def default_colormap():
            return cm.RdBu_r


def put_markers(markers):
    plt.scatter(markers[:, 0], markers[:, 1], c=markers[:, 2:5])


class MarkerDrawer(object):
    def __init__(self, markers):
        self.markers = markers

    @staticmethod
    def load(f):
        markers = list()
        for line in f:
            marker = MarkerDrawer.parse_line(line)
            if marker:
                markers.append(marker)

        return MarkerDrawer(markers)

    @staticmethod
    def load_from_file(path):
        with open(path) as f:
            return MarkerDrawer.load(f)

    @staticmethod
    def parse_line(line):
        line = line.strip()
        if line == "":
            return None
        if line.startswith("#"):
            return None
        data = re.split(r"\s+", line)
        if data[0] == "point":
            return ("point", (float(data[1]), float(data[2])), (float(data[3]), float(data[4]), float(data[5])))
        if data[0] in ["line", "arrow"]:
            return (
                data[0],
                (float(data[1]), float(data[2])),
                (float(data[3]), float(data[4])),
                (float(data[5]), float(data[6]), float(data[7])),
            )
        raise ValueError("cannot parse {}".format(line))

    def draw(self, ax):
        for marker in self.markers:
            if marker[0] == "arrow":
                (_, (x, y), (xdx, ydy), color) = marker
                ax.arrow(x, y, xdx - x, ydy - y, color=color, width=0.0001, length_includes_head=True)
            elif marker[0] == "line":
                (_, (x1, y1), (x2, y2), color) = marker
                ax.plot([x1, x2], [y1, y2], color=color)
            elif marker[0] == "point":
                (_, (x, y), color) = marker
                ax.scatter([x], [y], color=color, edgecolor="black")


if __name__ == "__main__":
    sys.exit(main())



import numpy as np

from homcloud.cubical_ext import CubeEncoder
import homcloud.visualize_3d as visualize_3d


class GeometryResolverBase(object):
    def boundary_cells(self, cell_indices):
        ret = set()
        for cell_index in cell_indices:
            for b in self.boundary_map(cell_index):
                if b in ret:
                    ret.remove(b)
                else:
                    ret.add(b)
        return ret

    def resolve_cells(self, cell_indices):
        return list(map(self.resolve_cell, cell_indices))

    def draw_cells(self, drawer, cells, color, **values):
        for cell in cells:
            self.draw_cell(drawer, cell, color, **values)

    def draw_boundary_cells(self, drawer, cells, color, **values):
        self.draw_cells(drawer, self.boundary_cells(cells), color, **values)


class CubicalResolver(GeometryResolverBase):
    def __init__(self, shape, index_to_cube, boundary_map):
        self.shape = shape
        self.cube_encoder = CubeEncoder(shape)
        self.index_to_cube = index_to_cube
        self.boundary_map = boundary_map

    @property
    def ndim(self):
        return len(self.shape)

    def resolve_cell(self, cell_index):
        return self.cube_encoder.decode_cube(self.index_to_cube[cell_index])

    def resolve_vertices(self, cell_indices):
        ret = set()

        def vertices_of_cube(cube_index):
            coord, nondeg = self.cube_encoder.decode_cube(self.index_to_cube[cube_index])

            def iter(k):
                if k == self.ndim:
                    ret.add(tuple(coord))
                    return
                if nondeg[k]:
                    iter(k + 1)
                    coord[k] += 1
                    iter(k + 1)
                    coord[k] -= 1
                else:
                    iter(k + 1)

            iter(0)
            return ret

        for cube_index in cell_indices:
            vertices_of_cube(cube_index)

        return list(ret)

    def boundary(self, cell_indices):
        return self.resolve_cells(self.boundary_cells(cell_indices))

    def boundary_vertices(self, cell_indices):
        return [list(vertex) for vertex in self.resolve_vertices(self.boundary_cells(cell_indices))]

    def centroid(self, cell_index):
        coord, nondeg = self.decode_index(cell_index)
        return np.array(coord) + np.array(nondeg) / 2.0

    def draw_cell(self, drawer, cell_index, color, **values):
        coords, nondeg = self.decode_index(cell_index)
        drawer.draw_cube(coords, nondeg, color, **values)

    def decode_index(self, index):
        return self.cube_encoder.decode_cube(self.index_to_cube[index])


class AbstractResolver(GeometryResolverBase):
    def __init__(self, index_to_symbol, boundary_map):
        self.index_to_symbol = index_to_symbol
        self.boundary_map = boundary_map

    def resolve_cell(self, cell_index):
        return self.index_to_symbol[cell_index]

    def boundary(self, cell_indices):
        return self.resolve_cells(self.boundary_cells(cell_indices))


class SimplicialResolver(GeometryResolverBase):
    def __init__(self, index_to_simplex, vertices, boundary_map):
        self.index_to_simplex = index_to_simplex
        self.vertices = vertices
        self.boundary_map = boundary_map

    def resolve_cell(self, cell_index):
        return [self.vertices[i] for i in self.index_to_simplex[cell_index]]

    def resolve_vertices(self, cell_indices):
        vertices = []
        for cell_index in cell_indices:
            for vertex in self.index_to_simplex[cell_index]:
                vertices.append(vertex)

        return [self.vertices[v] for v in set(vertices)]

    def boundary(self, cell_indices):
        return self.resolve_cells(self.boundary_cells(cell_indices))

    def boundary_vertices(self, cell_indices):
        return self.resolve_vertices(self.boundary_cells(cell_indices))

    def draw_boundary(self, drawer, cell_indices, color, **thvalues):
        for b in self.boundary_cells(cell_indices):
            drawer.draw_simplex(self.index_to_simplex[b], color, **thvalues)

    def build_paraview_drawer(self, n_colors, column_names):
        return visualize_3d.ParaViewSimplexDrawer(n_colors, self.vertices, column_names)

    def centroid(self, cell_index):
        return np.mean(self.resolve_cell(cell_index), axis=0)

    def draw_cell(self, drawer, cell_index, color, **values):
        drawer.draw_simplex(self.index_to_simplex[cell_index], color, **values)


class BitmapResolver(GeometryResolverBase):
    def __init__(self, index_to_pixel):
        self.index_to_pixel = index_to_pixel

    def resolve_cell(self, cell_index):
        return self.index_to_pixel[cell_index]

    def boundary(self, cell_indices):
        raise RuntimeError("boundary is not implemented in bitmap")


class RipsResolver(GeometryResolverBase):
    def __init__(self, vertices):
        self.vertices = vertices

    def resolve_graph_path(self, path):
        return [self.vertices[i] for i in path]



import argparse
import json
import os

from PIL import Image, ImageDraw
import numpy as np

from homcloud.version import __version__
from homcloud.view_index_pict import RE_PREDICATE_PAIR, OPERATORS, use_vectorized_histogram_mask
from homcloud.histogram import Histogram
from homcloud.argparse_common import parse_color
from homcloud.pdgm_format import PDGMReader


def main(args=None):
    args = args or argument_parser().parse_args()

    volume_info = load_volume_info(args.diagram, args.degree)
    nodes = [node for node in volume_info["nodes"].values() if node["death-time"] is not None]
    nodes = filter_nodes(nodes, args.filter)
    if use_vectorized_histogram_mask(args):
        nodes = filter_nodes_by_mask(nodes, args.vectorized_histogram_mask, args.histoinfo)
    image_drawer = Drawer.load_image(args.picture)
    image_drawer.setup_by_args(args)
    for node in nodes:
        image_drawer.draw_node_volume(node)
    for node in nodes:
        image_drawer.draw_node_birthdeath_pixel(node)

    if args.output:
        image_drawer.save(args.output)


def argument_parser():
    p = argparse.ArgumentParser(description="Show volume on pict2d")
    p.add_argument("-V", "--version", action="version", version=__version__)
    p.add_argument("-d", "--degree", type=int, required=True, help="degree of PH")
    p.add_argument("-f", "--filter", action="append", default=[], help='filters (ex: "lifetime > 5.0")')
    p.add_argument("-v", "--vectorized-histogram-mask", help="0-1 vector textfile for mask")
    p.add_argument("-H", "--histoinfo", help="vectorize histogram information" " (both -V and -H are required)")
    p.add_argument("-B", "--birth", default=False, action="store_true", help="plot birth pixels")
    p.add_argument("-D", "--death", default=False, action="store_true", help="plot death pixels")
    p.add_argument("--volume", default=False, action="store_true", help="plot volume pixels")
    p.add_argument("--birth-color", type=parse_color, default=(255, 0, 0), help="birth pixel color")
    p.add_argument("--death-color", type=parse_color, default=(0, 0, 255), help="death pixel color")
    p.add_argument("--volume-color", type=parse_color, default=(255, 0, 0), help="volume pixels color")
    p.add_argument("--alpha", type=float, default=0.5, help="alpha value for volume pixels")
    p.add_argument("-S", "--marker-size", default=1, type=int, help="marker size (default: 1)")
    p.add_argument("-o", "--output", help="output filername")
    p.add_argument("picture", help="input Picture file name")
    p.add_argument("diagram", help="persistence diagram file name (.p2mt)")
    return p


def load_volume_info(pdgm_path, degree):
    assert os.path.splitext(pdgm_path)[1] == ".pdgm"
    with PDGMReader.open(pdgm_path) as reader:
        assert degree in [0, reader.metadata["dim"] - 1]
        return reader.load_simple_chunk("bitmap_phtrees", degree)


def filter_nodes(nodes, filter_strings):
    predicates = [predicate_from_string(f) for f in filter_strings]
    return [node for node in nodes if all(pred(node) for pred in predicates)]


def predicate_from_string(string):
    m = RE_PREDICATE_PAIR.match(string)
    if not m:
        return None
    getter = {
        "birth": lambda node: node["birth-time"],
        "death": lambda node: node["death-time"],
        "lifetime": lambda node: node["death-time"] - node["birth-time"],
    }[m.group(1)]
    op = OPERATORS[m.group(2)]
    threshold = float(m.group(3))
    return lambda node: op(getter(node), threshold)


def filter_nodes_by_mask(nodes, mask_vector_path, histoinfo_path):
    print(mask_vector_path)
    vector = np.loadtxt(mask_vector_path).astype(bool)
    with open(histoinfo_path) as f:
        histoinfo = json.load(f)
    histogram = Histogram.reconstruct_from_vector(vector, histoinfo)
    return [node for node in nodes if histogram.value_at(node["birth-time"], node["death-time"])]


class Drawer(object):
    def __init__(self, image):
        self.image = image
        self.output_image = image.copy()
        self.draw = ImageDraw.Draw(self.output_image)
        self.draw_birth = self.draw_death = self.draw_volume = False
        self.birth_color = (255, 0, 0)
        self.death_color = (0, 0, 255)
        self.volume_color = (255, 0, 0)
        self.alpha = 0.5
        self.marker_size = 1

    @staticmethod
    def load_image(path):
        return Drawer(Image.open(path).convert("RGB"))

    def setup_by_args(self, args):
        self.draw_birth = args.birth
        self.draw_death = args.death
        self.draw_volume = args.volume
        self.birth_color = args.birth_color
        self.death_color = args.death_color
        self.volume_color = args.volume_color
        self.alpha = args.alpha
        self.marker_size = args.marker_size

    def draw_node_volume(self, node):
        if not self.draw_volume:
            return
        for pixel in node["volume"]:
            self.draw_volume_pixel(pixel)

    def draw_node_birthdeath_pixel(self, node):
        if self.draw_birth:
            self.draw_marker(node["birth-pixel"], self.birth_color)
        if self.draw_death:
            self.draw_marker(node["death-pixel"], self.death_color)

    def draw_marker(self, pixel, color):
        if pixel is None:
            return
        d = self.marker_size // 2
        for y in range(pixel[0] - d, pixel[0] + d + 1):
            for x in range(pixel[1] - d, pixel[1] + d + 1):
                self.draw.point((x, y), fill=color)

    def draw_volume_pixel(self, pixel):
        base_color = self.image.getpixel((pixel[1], pixel[0]))
        pixel_color = tuple(
            int(b * (1 - self.alpha) + c * self.alpha) for (b, c) in zip(base_color, self.volume_color)
        )
        self.draw.point((pixel[1], pixel[0]), fill=pixel_color)

    def save(self, path):
        self.output_image.save(path)


if __name__ == "__main__":
    main()



# You should run this script in the same directory of this file.

import os
import platform
import re

from setuptools import setup, Extension

import numpy as np

with open(os.path.join(os.path.dirname(__file__), "homcloud", "version.py")) as f:
    for line in f:
        m = re.match(r"__version__ = \"([\d.a-z]+)\"", line)
        if m:
            __version__ = m.group(1)


with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
    

if re.match(r'^MSC', platform.python_compiler()):
    CPPLANGOPTS = []
else:
    CPPLANGOPTS = ["-std=c++14", "-Wno-unknown-pragmas"]

LIBS = ["mpfr", "gmp"]

if os.environ.get("HOMCLOUD_BUILD_WITH_OPENMP", "0") == "1":
    openmp_compile_args = ["-fopenmp", "-DHOMCLOUD_OPENMP"]
    openmp_link_args = ["-fopenmp"]
else:
    openmp_compile_args = []
    openmp_link_args = []

PHAT_INCLUDE_DIR = "ext/external/phat/include"
MSGPACK_INCLUDE_DIR = "ext/external/msgpack-c/include"
HERA_INCLUDE_DIR = "ext/external/hera/include"


EXT_MODULES = [
    Extension("homcloud.modp_reduction_ext",
              sources=["ext/modp_reduction_ext.cc"],
              extra_compile_args=(CPPLANGOPTS + ["-DPYTHON"]),
              depends=[]),
    Extension("homcloud.alpha_shape3",
              include_dirs=[np.get_include()],
              libraries=LIBS,
              extra_compile_args=CPPLANGOPTS,
              define_macros=[("CGAL_HEADER_ONLY", None)],
              sources=["ext/alpha_shape3.cc"],
              depends=["ext/alpha_shape_common.h",
                       "ext/homcloud_common.h"]),
    Extension("homcloud.alpha_shape2",
              include_dirs=[np.get_include()],
              libraries=LIBS,
              extra_compile_args=CPPLANGOPTS,
              define_macros=[("CGAL_HEADER_ONLY", None)],
              sources=["ext/alpha_shape2.cc"],
              depends=["ext/alpha_shape_common.h",
                       "ext/homcloud_common.h"]),
    Extension("homcloud.periodic_alpha_shape3",
              include_dirs=[np.get_include()],
              libraries=LIBS,
              extra_compile_args=CPPLANGOPTS,
              define_macros=[("CGAL_HEADER_ONLY", None)],
              sources=["ext/periodic_alpha_shape3.cc"],
              depends=["ext/alpha_shape_common.h",
                       "ext/homcloud_common.h"]),
    Extension("homcloud.pict_tree",
              include_dirs=[np.get_include()],
              extra_compile_args=CPPLANGOPTS,
              sources=["ext/pict_tree.cc"],
              depends=["ext/homcloud_common.h"]),
    Extension("homcloud.cubical_ext",
              include_dirs=[np.get_include(), PHAT_INCLUDE_DIR, MSGPACK_INCLUDE_DIR],
              extra_compile_args=CPPLANGOPTS,
              sources=["ext/cubical_ext.cc"],
              depends=["ext/phat_ext.h"]),
    Extension("homcloud.phat_ext",
              include_dirs=[np.get_include(), PHAT_INCLUDE_DIR, MSGPACK_INCLUDE_DIR],
              extra_compile_args=(CPPLANGOPTS +
                                  openmp_compile_args),
              extra_link_args=openmp_link_args,
              sources=["ext/phat.cc"],
              depends=["ext/phat_ext.h"]),
    Extension("homcloud.distance_transform_ext",
              include_dirs=[np.get_include()],
              extra_compile_args=CPPLANGOPTS,
              sources=["ext/distance_transform_ext.cc"],
              depends=["ext/homcloud_common.h"]),
    Extension("homcloud.int_reduction_ext",
              extra_compile_args=CPPLANGOPTS,
              sources=["ext/int_reduction_ext.cc"],
              depends=[]),
    Extension("homcloud.homccube",
              include_dirs=[np.get_include()],
              extra_compile_args=CPPLANGOPTS,
              sources=["ext/homccube.cc"],
              depends=["ext/homcloud_common.h"]),
    Extension("homcloud.optimal_one_cycle_ext",
              include_dirs=[MSGPACK_INCLUDE_DIR],
              extra_compile_args=CPPLANGOPTS,
              sources=["ext/optimal_one_cycle_ext.cc"],
              depends=["ext/homcloud_common.h"]),
    Extension("homcloud.hera_bottleneck",
              include_dirs=[np.get_include(), HERA_INCLUDE_DIR],
              extra_compile_args=CPPLANGOPTS,
              sources=["ext/hera_bottleneck.cc"],
              depends=["ext/homcloud_common.h", "ext/homcloud_numpy.h", "ext/hera_common.h"]),
    Extension("homcloud.hera_wasserstein",
              extra_compile_args=CPPLANGOPTS,
              include_dirs=[np.get_include(), HERA_INCLUDE_DIR],
              sources=["ext/hera_wasserstein.cc"],
              depends=["ext/homcloud_common.h", "ext/homcloud_numpy.h", "ext/hera_common.h"]),
]


ENTRY_POINTS = {
    'console_scripts': [
        "homcloud-pc-alpha = homcloud.pc_alpha:main",
        "homcloud-pict-binarize-nd = homcloud.pict.binarize_nd:main",
        "homcloud-pict-pixel-levelset-nd = homcloud.pict.pixel_levelset_nd:main",
        "homcloud-pict-tree = homcloud.pict.tree:main",
        "homcloud-dump-diagram = homcloud.dump_diagram:main",
        "homcloud-plot-PD = homcloud.plot_PD:main",
        "homcloud-plot-PD-gui = homcloud.plot_PD_gui:main",
        "homcloud-vectorize-PD = homcloud.vectorize_PD:main",
        "homcloud-view-index-pict = homcloud.view_index_pict:main",
        "homcloud-view-vectorized-PD = homcloud.view_vectorized_PD:main",
        "homcloud-plot-PD-slice = homcloud.plot_PD_slice:main",
        "homcloud-pict-show-volume-2d = homcloud.pict.show_volume_2d:main",
        "homcloud-pict-pict3d-vtk = homcloud.pict.pict3d_vtk:main",
        "homcloud-pict-slice3d = homcloud.pict.slice3d:main",
        "homcloud-pict-to-npy = homcloud.pict.to_npy:main",
        "homcloud-rips = homcloud.rips:main",
        "homcloud-abstract-filtration = homcloud.abstract_filtration:main",
        "homcloud-build-phtrees = homcloud.build_phtrees:main",
        "homcloud-phtrees = homcloud.phtrees:main",
        "homcloud-optvol = homcloud.optvol:main",
    ],
}

    
setup(
    name="homcloud",
    version=__version__,
    description="HomCloud, persistent homology based data analysis package",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Ippei Obayashi",
    author_email="i.obayashi@okayama-u.ac.jp",
    url="http://homcloud.dev",
    ext_modules=EXT_MODULES,
    packages=[
        "homcloud",
        "homcloud.histopixels",
        "homcloud.pict",
        "homcloud.geometry",
        "homcloud.interface",
    ],
    entry_points=ENTRY_POINTS,
    install_requires=[
        "numpy",
        "matplotlib",
        "scipy",
        "scikit-learn",
        "msgpack",
        "Pillow",
        "pulp",
        "forwardable",
        "imageio",
        "Cython",
        "ripser",
        "cached-property",
        "plotly",
        "pyvista",
    ],
    setup_requires=["pytest-runner"],
    tests_require=[
        "pytest>=3.0",
        "pytest-mock",
        "pytest-lazy-fixture",
        "pytest-qt",
    ],
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Intended Audience :: Science/Research",
    ],
)



import numpy as np

from scipy.ndimage import distance_transform_cdt, distance_transform_edt
from sklearn.neighbors import BallTree
from homcloud.compat import DistanceMetric
from scipy.ndimage import binary_erosion

from homcloud.distance_transform_ext import positive_distance_transform_manhattan_with_mask
from homcloud.pict.utils import nd_indices


def distance_transform(pict, metric, signed=False, periodicity=None, obstacle=None, **opts):
    if obstacle is not None:
        check_arguments_for_obstacle(metric, pict, obstacle)

    if periodicity:
        pict = periodic_extend(pict, periodicity)
        if obstacle:
            obstacle = periodic_extend(pict, periodicity)
        distances = distance_transform(pict, metric, signed, None, obstacle, **opts)
        return periodic_shrink(distances, periodicity)

    npict = ~pict

    if signed:
        ret = -distance_transform(npict, metric, False, periodicity, None, **opts)
    else:
        ret = np.zeros_like(pict, dtype=float)

    if metric == "manhattan" and obstacle is not None:
        positive_distance_transform_manhattan_with_mask(pict, obstacle, ret)
    elif metric == "manhattan":
        ret[npict] = distance_transform_cdt(npict, "taxicab")[npict]
    elif metric == "chebyshev":
        ret[npict] = distance_transform_cdt(npict, "chessboard")[npict]
    elif metric == "euclidean":
        ret[npict] = distance_transform_edt(npict)[npict]
    elif metric == "mahalanobis":
        ret[npict] = distance_transform_mahalanobis(npict, **opts)[npict]
    else:
        raise ValueError("Unknown metric: {}".format(metric))

    return ret


def distance_transform_mahalanobis(npict, VI=None, V=None):
    if VI is not None:
        metric = DistanceMetric.get_metric("mahalanobis", VI=VI)
    elif V is not None:
        metric = DistanceMetric.get_metric("mahalanobis", V=V)

    ret = np.zeros_like(npict, dtype=float)
    balltree = BallTree(boundary_pixels(~npict), leaf_size=30, metric=metric)
    empty_pixels = pixel_positions(npict)

    def compute_distances(targets):
        (d, _) = balltree.query(targets, k=1)
        return d[:, 0]

    np.place(ret, npict, compute_distances(empty_pixels))

    return ret


def check_arguments_for_obstacle(metric, pict, obstacle):
    if metric != "manhattan":
        raise RuntimeError("obstacle is avaiable only with manhattan metric")
    if np.any(pict & obstacle):
        raise RuntimeError("obstacle must not have intersection with input image")


def pixel_positions(pict):
    """Return indices of True in 2D boolean ndarray pict"""
    indices = nd_indices(pict.shape)
    return indices[pict.flatten()]


def boundary_pixels(pict):
    return pixel_positions(np.logical_xor(pict, binary_erosion(pict, iterations=1, border_value=1)))


def periodic_extend(pict, periodicity):
    for nth, p in enumerate(periodicity):
        if p:
            pict = np.concatenate((pict, pict, pict), axis=nth)
    return pict


def periodic_shrink(pict, periodicity):
    for nth, p in enumerate(periodicity):
        if p:
            pict = np.split(pict, 3, axis=nth)[1]
    return pict



from forwardable import forwardable
import matplotlib.pyplot as plt

import homcloud.plot_PD as plot_PD
import homcloud.histogram as homhisto
import homcloud.vectorize_PD as vectorize_PD
import homcloud.plot_PD_slice as plot_PD_slice
import homcloud.pdgm as pdgm


@forwardable()
class HistoSpec(object):
    """
    This class represents a 2D mesh on the plane of birth-death pairs.

    This class is useful to compute histograms from many diagrams
    with the same setting.

    Args:
        x_range (tuple(float, float) or None): The lower and upper range
            of the bins on x-axis.
        y_range (int): The number of bins on x-axis.
        y_range (tuple(float, float) or None): The lower and upper range
            of the bins on y-axis. Same as `x_range` if None is given.
        y_bins (int or None): The number of bins on y-axis.
            Same as `x_bins` if None is given.
        superlevel (bool): This should be True if your PDs come
            from superlevel filtrations, otherwise this should be False.

    Examples:
        >>> import homcloud.interface as hc
        >>> import matplotlib.pyplot as plt
        >>> pc = np.array([[0, 0, 0], [8, 0, 0], [5, 6, 0], [4, 2, 6],])
        >>> pd = hc.PDList.from_alpha_filtration(pc).dth_diagram(1)
        >>> histospec = hc.HistoSpec((0, 20), 128)
        >>> histogram = histospec.pd_histogram(pd)
        >>> histogram.plot(colorbar={"type": "log"}, title="1st PD")
        >>> plt.show()
        -> The histogram is displayed.

    """

    def __init__(self, x_range, xbins, y_range=None, ybins=None, pd=None, superlevel=False):
        rulers = homhisto.Ruler.create_xy_rulers(x_range, xbins, y_range, ybins, pd)
        self.spec = homhisto.HistoSpec.from_rulers(rulers[0], rulers[1], superlevel)
        # self.rulers = rulers
        # self.sign_flipped = superlevel
        self._histoinfo = None

    def_delegators("spec", "sign_flipped")

    @classmethod
    def from_histoinfo(cls, histoinfo):
        new_mesh = cls(
            (histoinfo["x-edges"][0], histoinfo["x-edges"][-1]),
            len(histoinfo["x-edges"]) - 1,
            (histoinfo["y-edges"][0], histoinfo["y-edges"][-1]),
            len(histoinfo["y-edges"]) - 1,
        )
        new_mesh._histoinfo = histoinfo
        return new_mesh

    def pd_histogram(self, pd):
        """Constructs a 2D histogram of `pd`.

        Args:
            pd (:class:`PD`): The diagram.

        Returns:
            :class:`Histogram`: The histogram.
        """
        return Histogram(self.spec.histogram_from_diagram(pd))

    def histogram_from_vector(self, vector):
        """
        Construct a histogram from a vector.

        The histogram is constructed by the rule of persistence image method.

        Args:
            vector (numpy.ndarray): A vector.

        Returns:
            :class:`Histogram`: The histogram.
        """
        return Histogram(self.spec.histogram_from_vector(vector))

    def mask_from_vector(self, vector):
        """
        Construct a mask histogram from a boolean vector.

        The histogram is constructed by the rule of persistence image method.

        Args:
            vector (numpy.ndarray): A boolean vector.

        Returns:
            :class:`MaskHistogram`: The "histogram" whose values in bins are
                True or False.
        """
        return MaskHistogram(self.spec.histogram_from_vector(vector, homhisto.BinaryHistogram))

    def mask_from_2darray(self, array):
        return MaskHistogram(homhisto.BinaryHistogram(array, self.spec))

    def vector_size(self):
        """
        Return the size of the vector generated by self.

        Returns:
            int: The size of the vector
        """
        return self.spec.vector_size()


Mesh = HistoSpec


class PIVectorizeSpec(HistoSpec):
    """
    This class represents a 2D mesh on the plane of birth-death pairs
    with information about vectorization by persistence image (PI) method.

    You can construct PI vectors from diagrams.

    Args:
        x_range (tuple[float, float] or None): The lower and upper range
            of the bins on x-axis.
        xbins (int): The number of bins on x-axis.
        y_range (tuple[float, float] or None): The lower and upper range
            of the bins on y-axis. Same as `x_range` if None is given.
        ybins (int or None): The number of bins on y-axis.
            Same as `x_bins` if None is given.
        weight (tuple or string or Callable[[float, float], float]):
            The information of the weight function.

            You can use one of the followings:

            * "none": A constant weight function
            * Callable[[float, float], float]: Any callable object that computes weights.
              The first and second arguments are birth and death times of a pair
              and the return value is a positive weight value.
            * ("atan", `c`, `p`): An arctangent weight function,
              `c` and `p` should be floats. The weight function
              is atan(c * (death - birth)**p) .
            * ("linear", `a`): A linear weight function. `a` should be
              a float. The weight function is a*(death - birth) .
        sigma (float): The standard deviation for the Gaussian distribution
            used by PI.
        superlevel (bool): This should be True if your PDs come
            from superlevel filtrations, otherwise this should be False.
            Internally, This parameter should be True if the PD has
            birth-death pairs with birth > death, and otherwise
            the parameter should be False.
    """

    def __init__(self, x_range, xbins, y_range=None, ybins=None, weight=None, sigma=None, superlevel=False):
        super().__init__(x_range, xbins, y_range, ybins, None, superlevel)
        self.weight = self.weight_function(weight)
        self.sigma = sigma

    @staticmethod
    def weight_function(params):
        if params == "none":
            return lambda b, d: 1.0
        if isinstance(params, tuple):
            if params[0] == "atan":
                return vectorize_PD.atan_weight_function(params[1], params[2])
            if params[0] == "linear":
                return vectorize_PD.linear_weight_function(1.0 / params[1])
            raise ValueError("Unknown weight type: {}".format(params[0]))
        if callable(params):
            return params

    def vectorize(self, pd):
        """Vectroize `pd`.

        Args:
            pd (:class:`PD`): A persistence diagram.

        Returns:
            numpy.ndarray: The vectorized diagram.
        """
        assert pd.sign_flipped == self.sign_flipped
        histo = self.spec.histogram_from_diagram(pd)
        histo.apply_weight(self.weight)
        histo.apply_gaussian_filter(self.sigma)
        return histo.vectorize()

    def vectorize_pair(self, pair):
        """Vectorize a PD with a signle pair.

        Args:
            pd (:class:`Pair` or tuple[double, double]): A birth-death pair

        Returns:
            numpy.ndarray: The vectorized diagram.
        """
        birth_time, death_time = tuple(pair)
        return self.vectorize(pdgm.SimplePDGM(0, [birth_time], [death_time]))


PIVectorizerMesh = PIVectorizeSpec


@forwardable()
class Histogram(object):
    """The class represents a histogram on birth-death plane.

    Methods:
        x_range()
            Returns:
                (tuple[int, int]): The lower and upper range of x-axis.

        x_bins()
            Returns:
                (int): The number of bins on x-axis.

        y_range()
            Returns:
                (tuple[int, int]): The lower and upper range of y-axis.
        y_bins()
            Returns:
                (int): The number of bins on y-axis.

    Attributes:
        xedges (numpy.ndarray[x_bins + 1]): The edges of bins on x-axis
            in ascending order.
        yedges (numpy.ndarray[y_bins + 1]): The edges of bins on y-axis
            in ascending order.

    """

    def __init__(self, orig):
        self.orig = orig

    @property
    def values(self):
        """
        (numpy.ndarary, shape (x_bins, y_bins)): 2-dimensional array of values
        in the bins.
        """
        return self.orig.values

    @values.setter
    def values(self, values):
        self.orig.values = values

    def_delegators("orig", "x_range,y_range,x_bins,y_bins,xedges,yedges")

    def plot(
        self,
        colorbar={},
        style="colorhistogram",
        title="",
        unit_name=None,
        font_size=None,
        aspect="equal",
        ax=None,
        levels=None,
        plot_ess=False,
    ):
        """Plot a histogram by matplotlib.

        Args:
            colorbar (dict): The specification of the histogram colors and
                the colorbar.

                The following fields are available for this dictionary.

                * "type" - The name of colorbar type. One of the followings
                  is available. The default type is "linear".

                  * "linear" - linear scale
                  * "log" - log scale
                  * "loglog" - log(log(n+1)+1)
                  * "linear-midpoint" - linear scale with midpoint.
                    You should specify the value of midpoint
                    by the "midpoint" field.
                  * "power" - n^p. You should specify p by "p" field.
                * "min" - The minimum of the colorbar. If this value is not
                  specified, the value is determined by the minimum of
                  histogram values.
                * "max" - The maximum of the colorbar. If this value is not
                  specified, the value is determined by the maximum of
                  histogram values.
                * "colormap" - matplotlib's colormap name

            style (string): The plotting style. "colorhistogram" or "contour"
                is available.
            title (string): The title of the output figure.
            unit_name (string or None): The unit name of the x-axis
                and y-axis.
            font_size (float or None): The font size.
                The font size is automatically determined if None is given.
            aspect (string): The X-Y aspect. "equal" or "auto" is available.
            ax (matplotlib.axes.Axes or None):
                The axes to be plotted on. By default (if None),
                `matplotlib.pyplot.subplot()` is used.
            levels (list of floats or None):
                The levels for coutour plotting. This argument is used
                only if `style` is "contour".
            plot_ess (bool): Essential pairs are plotted if True.

        Example:
            >>> import matplotlib.pyplot as plt
            >>>    :
            >>> histogram.plot(colorbar={"type": "log", "max": 100},
            >>>                title="Title string")
            >>> plt.show()  # To show the histogram on your display
            >>> plt.savefig("histogram.png")  # To save the histogram figure.
        """
        plotter = plot_PD.PDPlotter.find_plotter(style)(
            self.orig, self._zspec(colorbar), plot_PD.AuxPlotInfo(title, unit_name, font_size, aspect, plot_ess)
        )
        if levels is not None:
            plotter.levels = levels

        plotter.plot(*self._fig_ax(ax))

    @staticmethod
    def _zspec(colorbar):
        ztype = colorbar.get("type", "linear")
        vmax = colorbar.get("max")
        vmin = colorbar.get("min")
        if "colormap" in colorbar:
            colormap = plt.get_cmap(colorbar["colormap"])
        else:
            colormap = None
        if ztype == "linear":
            return plot_PD.ZSpec.Linear(vmax, vmin, colormap)
        elif ztype == "log":
            return plot_PD.ZSpec.Log(vmax, vmin, colormap)
        elif ztype == "loglog":
            return plot_PD.ZSpec.LogLog(vmax, vmin, colormap)
        elif ztype == "linear-midpoint":
            return plot_PD.ZSpec.LinearMidPoint(colorbar["midpoint"], vmax, vmin, colormap)
        elif ztype == "power":
            return plot_PD.ZSpec.Power(colorbar["p"], vmax, vmin, colormap)
        else:
            raise RuntimeError("unknown colorbar type: {}").format(ztype)

    @staticmethod
    def _fig_ax(ax):
        if ax is None:
            return plt.subplots()
        else:
            return ax.get_figure(), ax


class MaskHistogram(Histogram):
    """
    The class represents a histogram on birth-death plane whose values
    are booleans.

    This class is helpful to pick up all birth-death pairs in an area.
    meth:`HistoSpec.mask_from_vector` and meth:`HistoSpec.mask_from_2darray` is
    available for this purpose.
    """

    def filter_pairs(self, pairs):
        """
        Returns all pairs in the area of bins whose values are True.

        Args:
            pairs (sequence of :class:`Pair`): Pairs to be filtered.

        Returns:
            list of :class:`Pair`: The filtered pairs.

        Notes:
            In fact, this method can filter the list of :class:`OptimalVolume`
            by their birth and death times.
        """
        return [pair for pair in pairs if self.orig.value_at(pair.birth_time(), pair.death_time())]

    def filter_pd(self, pd):
        """
        Returns all pairs of the persistence diagram
        in the area of bins whose values are True.

        Args:
            pd (:class:`PD`): A persistence diagram which has
               pairs to be filtered.

        Returns:
            list of :class:`Pair`: The filtered pairs.
        """

        return self.filter_pairs(pd.pairs())


class SliceHistogram(object):
    """
    This class represents a 1D histogram of birth-death pairs
    in the thin strip on a persistence diagram.

    You can create an instance by :meth:`PD.slice_histogram`.

    Attributes:
       values (numpy.ndarary): The histogram values.
       edges (nump.ndarray): The histogram bin edges.
       x1 (float): The x-coordinate of the one end of the strip.
       y1 (float): The y-coordinate of the one end of the strip.
       x2 (float): The x-coordinate of the another end of the strip.
       y2 (float): The y-coordinate of the another end of the strip.
    """

    def __init__(self, values, edges, x1, y1, x2, y2):
        self.values = values
        self.edges = edges
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2

    def plot(self, left_label=None, right_label=None, logscale=False, ax=None):
        """
        Plots the historam.

        Args:
            left_label (string): The label text at (x1, y1)
            right_label (string): The label text at (x2, y2)
            logscale (bool): Linear scale is used if False, and
               Log scale is used if True.
            ax (matplotlib.axes.Axes or None):
                The axes to be plotted on. By default (if None),
                `matplotlib.pyplot.gca()` is used.
        """
        ax = ax or plt.gca()
        width = self.edges[1] - self.edges[0]
        ax.bar(self.edges[:-1], self.values, width=width, align="edge", log=logscale)
        ax.set_xlim(-width / 2, 1 + width / 2)
        ax.set_xticks([0, 1])
        ax.set_xticklabels(
            [
                plot_PD_slice.construct_label(left_label, self.x1, self.y1),
                plot_PD_slice.construct_label(right_label, self.x2, self.y2),
            ]
        )



from cached_property import cached_property

from homcloud.base_filtration import Filtration
from homcloud.pdgm_format import PDGMWriter, BinaryChunk

import homcloud.phat_ext as phat


class SimplicialFiltration(Filtration):
    def __init__(self, simplicial_function, vertex_symbols, save_boundary_map):
        self.simplicial_function = self.normalize_simplicial_function(simplicial_function)
        s = sorted((level, len(simplex), simplex) for (simplex, level) in self.simplicial_function.items())
        self.index_to_simplex = [simplex for (_, _, simplex) in s]
        self.index_to_level = [level for (level, _, _) in s]
        self.simplex_to_index = {simplex: index for (index, (_, _, simplex)) in enumerate(s)}
        self.vertex_symbols = vertex_symbols or self.default_vertex_symbols()
        self.save_boundary_map = save_boundary_map

    @staticmethod
    def normalize_simplicial_function(fun):
        retval = dict()
        for simplex, level in fun.items():
            simplex = tuple(sorted(simplex))
            if simplex in retval:
                raise (ValueError("duplicated simplex"))

            retval[simplex] = level

        return retval

    def default_vertex_symbols(self):
        return [str(i) for i in range(self.num_vertices)]

    @cached_property
    def num_vertices(self):
        return max(k[0] for k in self.index_to_simplex if len(k) == 1) + 1

    def build_phat_matrix(self):
        matrix = phat.Matrix(len(self.index_to_simplex), self.boundary_map_style())
        for index, simplex in enumerate(self.index_to_simplex):
            boundary = self.boundary(simplex)
            self.check_boundary(boundary, index)
            matrix.set_dim_col(index, self.simplex_dim(simplex), boundary)
        return matrix

    def boundary_map_style(self):
        return "simplicial" if self.save_boundary_map else "none"

    @staticmethod
    def simplex_dim(simplex):
        return len(simplex) - 1

    def boundary(self, simplex):
        def kth(k):
            return self.simplex_to_index[tuple(simplex[:k] + simplex[k + 1 :])]

        if len(simplex) == 1:
            return []
        else:
            return [kth(k) for k in range(len(simplex))]

    @staticmethod
    def check_boundary(boundary, index):
        for i in boundary:
            assert i < index

    def compute_pdgm(self, f, algorithm=None, output_suppl_info=True):
        writer = PDGMWriter(f, "simplicial", self.dim)
        pairs, boundary_map_byte_sequence = self.compute_pairs(algorithm)
        writer.save_pairs(pairs, self.index_to_level, output_suppl_info)

        if output_suppl_info:
            writer.append_simple_chunk("index_to_level", self.index_to_level)
            writer.append_simple_chunk("index_to_simplex", self.index_to_simplex)
            writer.append_simple_chunk("vertex_symbols", self.vertex_symbols)

        if self.save_boundary_map:
            writer.append_chunk(BinaryChunk("boundary_map", boundary_map_byte_sequence))

        writer.write()

    @property
    def dim(self):
        return max(len(simplex) for simplex in self.index_to_simplex) - 1

    @staticmethod
    def favorite_algorithm():
        return "phat-twist"



"""HomCloud python interface

This module provides the interface to HomCloud from python program.

The API document is written by sphinx with Napolen.
http://www.sphinx-doc.org/ja/stable/ext/napoleon.html
"""
from itertools import chain
import subprocess
from tempfile import TemporaryDirectory

from PIL import Image
import numpy as np

import homcloud.pict.utils as pict_utils
import homcloud.view_index_pict as view_index_pict
import homcloud.pict.slice3d as pict_slice3d

from homcloud.license import LICENSE_TERMS

__all__ = [
    "LICENSE_TERMS",  # .homcloud.licence
    "HomCloudError",
    "VolumeNotFound",  # .exceptions
    "PDList",
    "PD",
    "Pair",  # .pd
    "HistoSpec",
    "Mesh",
    "PIVectorizeSpec",
    "PIVectorizerMesh",
    "Histogram",
    "MaskHistogram",
    "SliceHistogram",  # .histogram
    "Volume",
    "OptimalVolume",
    "EigenVolume",
    "StableVolume",
    "StableSubvolume",  # .optimal_volume
    "PHTrees",  # .phtrees
    "BitmapPHTrees",  # .bitmap_phtrees
    "distance_transform",  # .distance_transform
    "BitmapOptimal1Cycle",  # .bitmap_optimal_1_cycle
    "GraphOptimal1Cycle",  # .graph_optimal_1_cycle
    "Optimal1Cycle",  # .optimal_1_cycle
    "example_data",
    "loadtxt_nd",
    "draw_volumes_on_2d_image",
    "draw_birthdeath_pixels_2d",
    "show_slice3d",
]
from .exceptions import *
from .pd import *
from .histogram import *
from .optimal_volume import *
from .phtrees import *
from .bitmap_phtrees import *
from .bitmap_optimal_1_cycle import *
from .distance_transform import *
from . import bitmap as bitmap
from . import distance


def example_data(name):
    """Returns example data.

    Returns the tetrahedron 3D pointcloud for name == "tetrahedron".

    Args:
        name: Name of the data

    Examples:
        >>> import homcloud.interface as hc
        >>> hc.example_data("tetrahedron")
        array([[ 0.,  0.,  0.],
               [ 8.,  0.,  0.],
               [ 5.,  6.,  0.],
               [ 4.,  2.,  6.]])
    """
    if name == "tetrahedron":
        return np.array(
            [
                [0.0, 0.0, 0.0],
                [8.0, 0.0, 0.0],
                [5.0, 6.0, 0.0],
                [4.0, 2.0, 6.0],
            ]
        )
    if name == "bitmap_01_5x5x5":
        return np.array(
            [
                [[0, 0, 1, 1, 1], [0, 0, 0, 0, 1], [0, 0, 0, 0, 1], [0, 0, 0, 0, 1], [0, 0, 0, 1, 1]],
                [[0, 0, 1, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 1, 1, 0]],
                [[0, 0, 1, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 1, 0, 0]],
                [[0, 0, 1, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 1, 0, 0]],
                [[1, 1, 1, 0, 0], [1, 0, 0, 0, 0], [1, 0, 0, 0, 0], [1, 0, 0, 0, 0], [1, 1, 1, 0, 0]],
            ],
            dtype=bool,
        )
    if name == "bitmap_levels_5x5":
        return np.array(
            [
                [0, 0, 1, 1, 1],
                [0, 4, 0, 5, 1],
                [0, 8, 0, 3, 1],
                [0, 0, 4, 7, 1],
                [0, 0, 2, 1, 1],
            ],
            dtype=int,
        )

    raise ValueError("Unknown example name: %s" % name)


def loadtxt_nd(path):
    """
    Load a n-dimensional data from text.

    The format of the text file is as follows.

    * First line represents the shape of data. For example, if the shape of your d
      ata is 200x230x250, first line should be `200 230 250`.
    * The following lines are floating point number values in x-fastest direction
    * A line starting with `#` is skipped as a comment
    * An empty line is also skipped.

    The following is an example::

        # 4x3x2 3D voxel data
        4 3 2

        1 2 3 4
        5 6 7 8
        9 10 11 12

        13 14 15 16
        17 18 19 20
        21 22 23 24

    Args:
        path (string): The path of the text file.
    """
    return pict_utils.load_nd_picture_from_text(path)


def draw_pixels_on_2d_image(image, pixels, color, alpha=1.0):
    for y, x in pixels:
        if x < 0 or y < 0 or x >= image.size[0] or y >= image.size[1]:
            continue
        blended_color = tuple((np.array(image.getpixel((x, y))) * (1 - alpha) + np.array(color) * alpha).astype(int))
        image.putpixel((x, y), blended_color)
    return image


def draw_volumes_on_2d_image(nodes, image, color, alpha=1.0, birth_position=None, death_position=None, marker_size=1):
    """
    Draws optimal volumes for bitmap filtration on an image.

    Args:
        nodes (list of :class:`BitmapPHTrees.Node`):
            The tree nodes to be drawn.
        image (string or numpy.ndarray or PIL.Image.Image):
            The base image data.

            * string: The image file whose name is `image` is used.
            * numpy.ndarray: 2D array is treated as grayscale image.
            * PIL.Image.Image: The image data

            If PIL.Image.Image object is given, the object is overwrited.
        color (tuple[int, int, int]): The color (RGB) of the volumes.
        alpha (float): The alpha value.
           The volume is drawn by using alpha blending.
        birth_position (tuple[int, int, int] or None): If not None,
           birth positions are drawn by the given color
        death_position (tuple[int, int, int] or None): If not None,
           death positions are drawn by the given color
        marker_size (int): The marker size of birth positions and death positions.
           1, 3, 5, ... are available.
    Returns:
        PIL.Image.Image: The image data which optimal volumes are drawn.
    """
    image = to_pil_image(image)

    def draw_marker(position, color):
        d = marker_size // 2
        pixels = [(position[0] + dy, position[1] + dx) for dx in range(-d, d + 1) for dy in range(-d, d + 1)]
        draw_pixels_on_2d_image(image, pixels, color)

    pixels = set(chain.from_iterable([map(tuple, node.volume()) for node in nodes]))
    draw_pixels_on_2d_image(image, pixels, color, alpha)

    if birth_position:
        for node in nodes:
            draw_marker(node.birth_position(), birth_position)
    if death_position:
        for node in nodes:
            draw_marker(node.death_position(), death_position)

    return image


def draw_birthdeath_pixels_2d(
    pairs,
    image,
    draw_birth=False,
    draw_death=False,
    draw_line=False,
    scale=1,
    marker_type="filled-diamond",
    marker_size=1,
    with_label=False,
    birth_color=(255, 0, 0),
    death_color=(0, 0, 255),
    line_color=(0, 255, 0),
):
    """
    Draw birth/death pixels on the given image.

    This function returns PIL.Image.Image object.
    Please see the `document of pillow
    <https://pillow.readthedocs.io/en/latest/reference/Image.html>`_
    to know how to treat this object.

    Args:
        pairs (list of :class:`Pair`): The birth-death pairs.
        image (string or numpy.ndarray or PIL.Image.Image):
            The image data.

            * string: The image file whose name is `image` is used.
            * numpy.ndarray: 2D array is treated as grayscale image.
            * PIL.Image.Image: The image data
        draw_birth (bool): Birth pixels are drawn if True.
        draw_death (bool): Death pixels are drawn if True.
        draw_line (bool): The lines connecting each birth pixels and
            death pixels are drawn.
        scale (int): Image scaling factor.
        marker_type (string): The type of the markers. You can choose
            one of the followings:

            * "filled-diamond"
            * "point"
            * "square"
            * "filled-square"
            * "circle"
            * "filled-circle"
        marker_size (int): The size of the marker.
        with_label (bool): Show birth and death times beside
            each birth/death pixel marker.
        birth_color (tuple[int, int, int]): The color of birth pixel markers.
        death_color (tuple[int, int, int]): The color of death pixel markers.
        line_color (tuple[int, int, int]): The color of lines.

    Returns:
        PIL.Image.Image: The image data which birth/death pixels are drawn.
    """
    image = to_pil_image(image)

    output_image, marker_drawer = view_index_pict.setup_images(image, scale, not with_label)
    marker_drawer.scale = scale
    marker_drawer.marker_size = marker_size
    marker_drawer.should_draw_birth_pixel = draw_birth
    marker_drawer.should_draw_death_pixel = draw_death
    marker_drawer.should_draw_line = draw_line
    marker_drawer.no_label = not with_label
    marker_drawer.birth_color = birth_color
    marker_drawer.death_color = death_color
    marker_drawer.line_color = line_color

    for pair in pairs:
        marker_drawer.draw_pair(pair)

    return output_image


def to_pil_image(image):
    if isinstance(image, str):
        return Image.open(image).convert("RGB")
    elif isinstance(image, np.ndarray):
        upper = np.max(image)
        lower = np.min(image)
        return Image.fromarray(((image - lower) / (upper - lower) * 250).astype(np.uint8)).convert("RGB")
    elif isinstance(image, Image.Image):
        return image
    else:
        raise RuntimeError("{} cannot be converted to PIL.Image".format(repr(image)))


def show_slice3d(volumes, slice=0, spacer=0, range=None, image_viewer="eog -n"):
    """
    Display slices of 3D bitmap data.

    Args:
        volumes (list of numpy.ndarray): multiple 3D bitmap data. These data are
            aligned horizontally.
        slice (int): The direction of slicing: 0, 1, or 2 for z, y, x direction.
        spacer (int): The number of pixels between horizontally aligned slices.
        range (None or tuple[int, int]): The range of the slices.
        image_viewer (str): Command line of the image viewer to see slices.
    """
    with TemporaryDirectory() as tmpdir:
        pict_slice3d.write_volume_slices(volumes, slice, spacer, range, tmpdir)
        subprocess.call("{} {}".format(image_viewer, tmpdir), shell=True)
